from abc import abstractmethod
from copy import copy, deepcopy
from itertools import chain
from os import linesep
from types import FunctionType
from typing import Tuple, List, Union, Optional

from .client_base import Table, KustoResponseBase, RetryConfig, ClientRequestProperties
from .enums import Order, Nulls, JoinKind, Distribution, BagExpansion
from .expressions import BooleanType, ExpressionType, AggregationExpression, _AssignmentBase, _AssignmentFromAggregationToColumn, _AssignmentToSingleColumn, \
    _AnyTypeColumn, BaseExpression, _AssignmentFromColumnToColumn, AnyExpression, _to_kql, _expression_to_type, BaseColumn, NumberType, OrderedType
# These seem like redundant imports, but without them typeguard is confused
# noinspection PyUnresolvedReferences
from .expressions import _DatetimeExpression, _TimespanExpression, _NumberExpression, _StringExpression, _BooleanExpression  # noqa: F401
from .functions import Functions as f
from .kql_converters import KQL
from .logger import _logger
from .type_utils import _KustoType, _typed_column, _plain_expression
from .udf import _stringify_python_func


class Query:
    _head: Optional['Query']
    _table: Optional[Table]
    _table_name: Optional[str]

    def __init__(self, head=None) -> None:
        """
        When building the query, try and use the query best practices:
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/best-practices
        """
        self._head = head if isinstance(head, Query) else None
        self._table = head if isinstance(head, Table) else None
        self._table_name = head if isinstance(head, str) else None

    def __add__(self, other: 'Query') -> 'Query':
        self_copy = deepcopy(self)
        other_copy = deepcopy(other)

        other_base = other_copy
        while other_base._head is not None:
            if other_base._head._head is None:
                break
            other_base = other_base._head
        other_base._head = self_copy
        return other_copy

    def __deepcopy__(self, memo) -> 'Query':
        new_object = copy(self)
        if self._head is not None:
            new_object._head = self._head.__deepcopy__(memo)
        return new_object

    def where(self, *predicates: BooleanType) -> 'Query':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/whereoperator

        Implicitly apply conjunction if multiple predicates are provided. You can use predicates which are calculated at runtime and result in boolean values, which are
        pre-processed: 'True' values ignored, and 'False' values cause all other predicates to be ignored. If the result of pre-processing is a single 'True' predicate, no 'where'
        clause will be generated.

        Best practices (`full list <https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/best-practices>`_):

        * Whenever possible, use time filters first.
        * Prefer filtering on table columns, and not on calculated columns.
        * Simplest terms first: If you have multiple clauses conjoined with and, put first the clauses that involve just one column.

        Warning: to apply a logical 'not', do not use the Python 'not' operator, it will simply produce a 'False' boolean value. Use either the `~` operator or `f.not_of()`.
        """
        filtered_predicates = []
        for predicate in predicates:
            if predicate is True:
                # This predicate has no effect on the outcome
                continue
            if predicate is False:
                # All other predicates have no effect on the outcome
                filtered_predicates = [False]
                break
            filtered_predicates.append(predicate)
        if len(filtered_predicates) == 0:
            # Do no generate 'where' clause
            return self
        return _WhereQuery(self, *filtered_predicates)

    def take(self, num_rows: int) -> '_TakeQuery':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/takeoperator
        """
        return _TakeQuery(self, num_rows)

    def limit(self, num_rows: int) -> '_LimitQuery':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/limitoperator
        """
        return _LimitQuery(self, num_rows)

    def sample(self, num_rows: int) -> '_SampleQuery':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/sampleoperator
        """
        return _SampleQuery(self, num_rows)

    def count(self) -> '_CountQuery':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/countoperator
        """
        return _CountQuery(self)

    def sort_by(self, col: OrderedType, order: Order = None, nulls: Nulls = None) -> '_SortQuery':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/sortoperator
        """
        return _SortQuery(self, col, order, nulls)

    def order_by(self, col: OrderedType, order: Order = None, nulls: Nulls = None) -> '_SortQuery':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/orderoperator
        """
        return self.sort_by(col, order, nulls)

    def top(self, num_rows: int, col: OrderedType, order: Order = None, nulls: Nulls = None) -> '_TopQuery':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/topoperator
        """
        return _TopQuery(self, num_rows, col, order, nulls)

    def join(self, query: 'Query', kind: JoinKind = None) -> '_JoinQuery':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/joinoperator

        Best practices (`full list <https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/best-practices>`_):

        * Select the table with the fewer rows to be the first one ("left" side).
        * Across clusters, run the query on the "right" side of the join, where most of the data is located.

        Best practices once `hint.strategy` is supported (see `here <https://github.com/Azure/pykusto/issues/104>`_):

        * When left side is small (up to ~100,000 records) and right side is large use `hint.strategy=broadcast`.
        * When both sides are too large	use `hint.strategy=shuffle`.
        """
        return _JoinQuery(self, query, kind)

    def project(self, *args: Union[_AssignmentBase, BaseExpression], **kwargs: ExpressionType) -> '_ProjectQuery':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/projectoperator
        """
        return _ProjectQuery(self, self._extract_assignments(*args, **kwargs))

    def project_rename(self, *args: _AssignmentFromColumnToColumn, **kwargs: BaseColumn) -> '_ProjectRenameQuery':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/projectrenameoperator
        """
        assignments: List[_AssignmentFromColumnToColumn] = list(args)
        for column_name, column in kwargs.items():
            assignments.append(_AssignmentFromColumnToColumn(_AnyTypeColumn(column_name), column))
        return _ProjectRenameQuery(self, assignments)

    def project_away(self, *columns: Union[BaseColumn, str]) -> '_ProjectAwayQuery':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/projectawayoperator
        """
        return _ProjectAwayQuery(self, columns)

    def distinct(self, *columns: BaseColumn) -> '_DistinctQuery':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/distinctoperator
        """
        return _DistinctQuery(self, columns)

    def distinct_all(self) -> '_DistinctQuery':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/distinctoperator
        """
        return _DistinctQuery(self, (_AnyTypeColumn(KQL("*")),))

    def extend(self, *args: Union[BaseExpression, _AssignmentBase], **kwargs: ExpressionType) -> '_ExtendQuery':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/extendoperator
        """
        return _ExtendQuery(self, *self._extract_assignments(*args, **kwargs))

    def summarize(self, *args: Union[AggregationExpression, _AssignmentFromAggregationToColumn],
                  **kwargs: AggregationExpression) -> '_SummarizeQuery':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/summarizeoperator
        """
        assignments: List[_AssignmentFromAggregationToColumn] = []
        for arg in args:
            if isinstance(arg, AggregationExpression):
                assignments.append(arg.assign_to())
            else:
                assert isinstance(arg, _AssignmentFromAggregationToColumn), "Invalid assignment"
                assignments.append(arg)
        for column_name, agg in kwargs.items():
            assignments.append(_AssignmentFromAggregationToColumn(_AnyTypeColumn(column_name), agg))
        return _SummarizeQuery(self, assignments)

    def mv_expand(
            self, *args: Union[BaseExpression, _AssignmentBase], bag_expansion: BagExpansion = None,
            with_item_index: BaseColumn = None, limit: int = None, **kwargs: ExpressionType
    ) -> '_MvExpandQuery':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/mvexpandoperator
        """
        assignments = self._extract_assignments(*args, **kwargs)
        if len(assignments) == 0:
            raise ValueError("Please specify one or more columns for mv-expand")
        return _MvExpandQuery(self, bag_expansion, with_item_index, limit, *assignments)

    def custom(self, custom_query: str) -> '_CustomQuery':
        return _CustomQuery(self, custom_query)

    def evaluate(self, plugin_name, *args: ExpressionType, distribution: Distribution = None) -> '_EvaluateQuery':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/evaluateoperator
        """
        return _EvaluateQuery(self, plugin_name, *args, distribution=distribution)

    def evaluate_udf(
            self, udf: FunctionType, extend: bool = True, distribution: Distribution = None, **type_specs: _KustoType
    ) -> '_EvaluateQuery':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/pythonplugin
        """
        return _EvaluateQuery(
            self, 'python',
            AnyExpression(KQL(f'typeof({("*, " if extend else "") + ", ".join(field_name + ":" + kusto_type.primary_name for field_name, kusto_type in type_specs.items())})')),
            _stringify_python_func(udf),
            distribution=distribution
        )

    def bag_unpack(self, col: BaseColumn, prefix: str = None) -> '_EvaluateQuery':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/bag-unpackplugin
        """
        if prefix is None:
            return _EvaluateQuery(self, 'bag_unpack', col)
        return _EvaluateQuery(self, 'bag_unpack', col, prefix)

    @abstractmethod
    def _compile(self) -> KQL:
        raise NotImplementedError()

    def _compile_all(self, use_full_table_name) -> KQL:
        if self._head is None:
            if self._table is None:
                if self._table_name is None:
                    return KQL("")
                else:
                    return KQL(self.get_table_name())
            else:
                table = self._table
                if use_full_table_name:
                    return table.to_query_format(fully_qualified=True)
                else:
                    return table.to_query_format()
        else:
            return KQL(f"{self._head._compile_all(use_full_table_name)} | {self._compile()}")

    def get_table(self) -> Optional[Table]:
        if self._head is None:
            return self._table
        else:
            return self._head.get_table()

    def get_table_name(self) -> Optional[str]:
        if self._head is None:
            return self._table_name
        else:
            return self._head.get_table_name()

    def render(self, use_full_table_name: bool = False) -> KQL:
        result = self._compile_all(use_full_table_name)
        _logger.debug("Complied query: " + result)
        return result

    def pretty_render(self, use_full_table_name: bool = False) -> KQL:
        kql = self.render(use_full_table_name)
        if kql is not None:
            kql = KQL(kql.replace(" |", linesep + "|"))
        return kql

    def execute(self, table: Table = None, properties: ClientRequestProperties = None, retry_config: RetryConfig = None) -> KustoResponseBase:
        if self.get_table() is None:
            if table is None:
                raise RuntimeError("No table supplied")
            rendered_query = table.to_query_format() + self.render()
        else:
            if table is not None:
                raise RuntimeError("This table is already bound to a query")
            table = self.get_table()
            rendered_query = self.render()

        _logger.debug("Running query: " + rendered_query)
        return table.execute(rendered_query, properties, retry_config)

    def to_dataframe(self, table: Table = None, properties: ClientRequestProperties = None, retry_config: RetryConfig = None):
        return self.execute(table, properties, retry_config).to_dataframe()

    @staticmethod
    def _extract_assignments(*args: Union[_AssignmentBase, BaseExpression], **kwargs: ExpressionType) -> List[_AssignmentBase]:
        assignments: List[_AssignmentBase] = []
        for arg in args:
            if isinstance(arg, BaseExpression):
                assignments.append(arg.assign_to())
            else:
                assignments.append(arg)
        for column_name, expression in kwargs.items():
            column_type = _expression_to_type(expression, _typed_column, _AnyTypeColumn)
            if isinstance(expression, BaseExpression):
                assignments.append(expression.assign_to(column_type(column_name)))
            else:
                expression_type = _expression_to_type(expression, _plain_expression, AnyExpression)
                assignments.append(expression_type(_to_kql(expression)).assign_to(column_type(column_name)))
        return assignments


class _ProjectQuery(Query):
    _assignments: List[_AssignmentBase]

    def __init__(self, head: 'Query', assignments: List[_AssignmentBase]) -> None:
        super().__init__(head)
        self._assignments = assignments

    def _compile(self) -> KQL:
        return KQL(f"project {', '.join(a.to_kql() for a in self._assignments)}")


class _ProjectRenameQuery(Query):
    _assignments: List[_AssignmentBase]

    def __init__(self, head: 'Query', assignments: List[_AssignmentFromColumnToColumn]) -> None:
        super().__init__(head)
        self._assignments = assignments

    def _compile(self) -> KQL:
        return KQL(f"project-rename {', '.join(a.to_kql() for a in self._assignments)}")


class _ProjectAwayQuery(Query):
    _columns: Tuple[BaseColumn, ...]

    def __init__(self, head: 'Query', columns: Tuple[Union[BaseColumn, str], ...]) -> None:
        super().__init__(head)
        self._columns = tuple(c if isinstance(c, BaseColumn) else _AnyTypeColumn(c) for c in columns)

    def _compile(self) -> KQL:
        return KQL(f"project-away {', '.join(_to_kql(c) for c in self._columns)}")


class _DistinctQuery(Query):
    _columns: Tuple[BaseColumn, ...]

    def __init__(self, head: 'Query', columns: Tuple[BaseColumn, ...]) -> None:
        super().__init__(head)
        self._columns = columns

    def sample(self, number_of_values: NumberType) -> '_SampleDistinctQuery':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/sampledistinctoperator
        """
        assert len(self._columns) == 1, "sample-distinct supports only one column"
        return _SampleDistinctQuery(self._head, self._columns[0], number_of_values)

    def top_hitters(self, number_of_values: NumberType) -> '_TopHittersQuery':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/tophittersoperator
        """
        assert len(self._columns) == 1, "top-hitters supports only one column"
        return _TopHittersQuery(self._head, self._columns[0], number_of_values)

    def _compile(self) -> KQL:
        return KQL(f"distinct {', '.join(c.kql for c in self._columns)}")


class _SampleDistinctQuery(Query):
    _number_of_values: NumberType
    _column: BaseColumn

    def __init__(self, head: 'Query', column: BaseColumn, number_of_values: NumberType) -> None:
        super().__init__(head)
        self._column = column
        self._number_of_values = number_of_values

    def _compile(self) -> KQL:
        return KQL(f"sample-distinct {_to_kql(self._number_of_values)} of {self._column.kql}")


class _TopHittersQuery(Query):
    _number_of_values: NumberType
    _column: BaseColumn
    _by_expression: Optional[NumberType]

    def __init__(self, head: 'Query', column: BaseColumn, number_of_values: NumberType, by_expression: Optional[NumberType] = None) -> None:
        super().__init__(head)
        self._column = column
        self._number_of_values = number_of_values
        self._by_expression = by_expression

    def by(self, by_expression: NumberType) -> '_TopHittersQuery':
        assert self._by_expression is None, "duplicate 'by' clause"
        return _TopHittersQuery(self._head, self._column, self._number_of_values, by_expression)

    def _compile(self) -> KQL:
        return KQL(f"top-hitters {_to_kql(self._number_of_values)} of {self._column.kql}{'' if self._by_expression is None else f' by {_to_kql(self._by_expression)}'}")


class _ExtendQuery(Query):
    _assignments: Tuple[_AssignmentBase, ...]

    def __init__(self, head: 'Query', *assignments: _AssignmentBase) -> None:
        super().__init__(head)
        self._assignments = assignments

    def _compile(self) -> KQL:
        return KQL(f"extend {', '.join(a.to_kql() for a in self._assignments)}")


class _WhereQuery(Query):
    _predicates: Tuple[BooleanType, ...]

    def __init__(self, head: Query, *predicates: BooleanType):
        super(_WhereQuery, self).__init__(head)
        self._predicates = predicates

    def _compile(self) -> KQL:
        if len(self._predicates) == 1:
            return KQL(f'where {_to_kql(self._predicates[0])}')
        return KQL(f'where {f.all_of(*self._predicates)}')


class _SingleNumberQuery(Query):
    _num_rows: int
    _query_name: str

    def __init__(self, head: Query, query_name: str, num_rows: int):
        super(_SingleNumberQuery, self).__init__(head)
        self._query_name = query_name
        self._num_rows = num_rows

    def _compile(self) -> KQL:
        return KQL(f'{self._query_name} {self._num_rows}')


class _TakeQuery(_SingleNumberQuery):
    _num_rows: int

    def __init__(self, head: Query, num_rows: int):
        super(_TakeQuery, self).__init__(head, 'take', num_rows)


class _LimitQuery(_SingleNumberQuery):
    _num_rows: int

    def __init__(self, head: Query, num_rows: int):
        super(_LimitQuery, self).__init__(head, 'limit', num_rows)


class _SampleQuery(_SingleNumberQuery):
    _num_rows: int

    def __init__(self, head: Query, num_rows: int):
        super(_SampleQuery, self).__init__(head, 'sample', num_rows)


class _CountQuery(Query):
    _num_rows: int

    def __init__(self, head: Query):
        super(_CountQuery, self).__init__(head)

    def _compile(self) -> KQL:
        return KQL('count')


class _OrderQueryBase(Query):
    class OrderSpec:
        col: OrderedType
        order: Optional[Order]
        nulls: Optional[Nulls]

        def __init__(self, col: OrderedType, order: Optional[Order], nulls: Optional[Nulls]):
            self.col = col
            self.order = order
            self.nulls = nulls

    _query_name: str
    _order_specs: List[OrderSpec]

    def __init__(self, head: Query, query_name: str, col: OrderedType, order: Optional[Order], nulls: Optional[Nulls]):
        super().__init__(head)
        self._query_name = query_name
        self._order_specs = []
        self.then_by(col, order, nulls)

    def then_by(self, col: OrderedType, order: Order = None, nulls: Nulls = None):
        self._order_specs.append(_OrderQueryBase.OrderSpec(col, order, nulls))
        return self

    @staticmethod
    def _compile_order_spec(order_spec: OrderSpec) -> str:
        res = str(order_spec.col.kql)
        if order_spec.order is not None:
            res += " " + str(order_spec.order.value)
        if order_spec.nulls is not None:
            res += " nulls " + str(order_spec.nulls.value)
        return res

    def _compile(self) -> KQL:
        return KQL(f'{self._query_name} by {", ".join([self._compile_order_spec(order_spec) for order_spec in self._order_specs])}')


class _SortQuery(_OrderQueryBase):
    def __init__(self, head: Query, col: OrderedType, order: Optional[Order], nulls: Optional[Nulls]):
        super(_SortQuery, self).__init__(head, "sort", col, order, nulls)


class _TopQuery(Query):
    _num_rows: int
    _order_spec: _OrderQueryBase.OrderSpec

    def __init__(self, head: Query, num_rows: int, col: OrderedType, order: Order, nulls: Nulls):
        super(_TopQuery, self).__init__(head)
        self._num_rows = num_rows
        self._order_spec = _OrderQueryBase.OrderSpec(col, order, nulls)

    def _compile(self) -> KQL:
        # noinspection PyProtectedMember
        return KQL(f'top {self._num_rows} by {_SortQuery._compile_order_spec(self._order_spec)}')


class JoinException(Exception):
    pass


class _JoinQuery(Query):
    _joined_query: Query
    _kind: JoinKind
    _on_attributes: Tuple[Union[Tuple[BaseColumn], Tuple[BaseColumn, BaseColumn]], ...]

    def __init__(self, head: Query, joined_query: Query, kind: Optional[JoinKind],
                 on_attributes: Tuple[Tuple[BaseColumn, ...], ...] = tuple()):
        super(_JoinQuery, self).__init__(head)
        self._joined_query = joined_query
        self._kind = kind
        self._on_attributes = on_attributes

    def on(self, *join_cols: Union[BaseColumn, Tuple[BaseColumn, BaseColumn]]) -> '_JoinQuery':
        for col in join_cols:
            self._inner_on_with_or_without_table(col)
        return self

    def _inner_on_with_or_without_table(self, col: Union[BaseColumn, Tuple[BaseColumn, BaseColumn]]) -> '_JoinQuery':
        if isinstance(col, BaseColumn):
            return self._inner_on(col)
        elif isinstance(col, tuple) and len(col) == 2 and isinstance(col[0], BaseColumn) and isinstance(col[1], BaseColumn):
            return self._inner_on_with_table(*col)
        else:
            # Re-apply coverage to this line once the corresponding test is re-enabled, after this is resolved: https://github.com/agronholm/typeguard/issues/159
            raise JoinException(  # pragma: no cover
                "A join argument could be a column, or a tuple of two columns corresponding to the input and join "
                f"tables column names. However, the join argument provided is {col} of type {type(col)}"
            )

    def _inner_on(self, col: BaseColumn) -> '_JoinQuery':
        self._on_attributes = self._on_attributes + ((col,),)
        return self

    def _inner_on_with_table(self, col1: BaseColumn, col2: BaseColumn) -> '_JoinQuery':
        self._on_attributes = self._on_attributes + ((col1, col2),)
        return self

    @staticmethod
    def _compile_on_attribute(attribute: Union[Tuple[BaseColumn], Tuple[BaseColumn, BaseColumn]]):
        assert len(attribute) in (1, 2)
        if len(attribute) == 1:
            return attribute[0].kql
        else:
            return f"$left.{attribute[0].kql}==$right.{attribute[1].kql}"

    def _assert_joined_query(self) -> None:
        if self._joined_query.get_table() is None:
            raise JoinException("The joined query must have a table")

    def _assert_on_attributes(self) -> None:
        if len(self._on_attributes) == 0:
            raise JoinException("A call to join() must be followed by a call to on()")

    def _compile(self) -> KQL:
        self._assert_joined_query()
        self._assert_on_attributes()

        return KQL(f'join {"" if self._kind is None else f"kind={self._kind.value}"} '
                   f'({self._joined_query.render(use_full_table_name=True)}) on '
                   f'{", ".join([self._compile_on_attribute(attr) for attr in self._on_attributes])}')


class _SummarizeQuery(Query):
    _assignments: List[_AssignmentFromAggregationToColumn]
    _by_columns: List[Union[BaseColumn, BaseExpression]]
    _by_assignments: List[_AssignmentToSingleColumn]

    def __init__(self, head: Query,
                 assignments: List[_AssignmentFromAggregationToColumn]):
        super(_SummarizeQuery, self).__init__(head)
        self._assignments = assignments
        self._by_columns = []
        self._by_assignments = []

    def by(self, *args: Union[_AssignmentToSingleColumn, BaseColumn, BaseExpression],
           **kwargs: BaseExpression):
        for arg in args:
            if isinstance(arg, BaseColumn) or isinstance(arg, BaseExpression):
                self._by_columns.append(arg)
            else:
                assert isinstance(arg, _AssignmentToSingleColumn), "Invalid assignment"
                self._by_assignments.append(arg)
        for column_name, group_exp in kwargs.items():
            self._by_assignments.append(_AssignmentToSingleColumn(_AnyTypeColumn(column_name), group_exp))
        return self

    def _compile(self) -> KQL:
        result = f"summarize {', '.join(a.to_kql() for a in self._assignments)}"
        if len(self._by_assignments) != 0 or len(self._by_columns) != 0:
            result += f' by {", ".join(chain((c.kql for c in self._by_columns), (a.to_kql() for a in self._by_assignments)))}'
        return KQL(result)


class _MvExpandQuery(Query):
    _assignments: Tuple[_AssignmentBase]
    _bag_expansion: Optional[BagExpansion]
    _with_item_index: Optional[BaseColumn]
    _limit: Optional[int]

    def __init__(self, head: Query, bag_expansion: Optional[BagExpansion], with_item_index: Optional[BaseColumn], limit: Optional[int], *assignments: _AssignmentBase):
        super(_MvExpandQuery, self).__init__(head)
        self._assignments = assignments
        self._bag_expansion = bag_expansion
        self._with_item_index = with_item_index
        self._limit = limit

    def _compile(self) -> KQL:
        res = "mv-expand "
        if self._bag_expansion is not None:
            res += f"bagexpansion={self._bag_expansion.value} "
        if self._with_item_index is not None:
            res += f"with_itemindex={self._with_item_index.kql} "
        res += ", ".join(a.to_kql() for a in self._assignments)
        if self._limit is not None:
            res += f" limit {self._limit}"
        return KQL(res)


class _CustomQuery(Query):
    _custom_query: str

    def __init__(self, head: Query, custom_query: str):
        super(_CustomQuery, self).__init__(head)
        self._custom_query = custom_query

    def _compile(self) -> KQL:
        return KQL(self._custom_query)


class _EvaluateQuery(Query):
    _plugin_name: str
    _args: Tuple[ExpressionType]
    _distribution: Distribution

    def __init__(self, head: Query, plugin_name: str, *args: ExpressionType, distribution: Distribution = None):
        super().__init__(head)
        self._plugin_name = plugin_name
        self._args = args
        self._distribution = distribution

    def _compile(self) -> KQL:
        return KQL(f'evaluate {"" if self._distribution is None else f"hint.distribution={self._distribution.value} "}'
                   f'{self._plugin_name}({", ".join(_to_kql(arg) for arg in self._args)})')



# flake8: noqa
# Allows importing externally-facing classes without specifying submodules.
# e.g. "from pykusto import PyKustoClient" instead of "from pykusto.client import PyKustoClient"
# Also allows for a convenient list of all externally-facing classes as the autocomplete of "from pykusto import "
# "import *" does not import names which start with an underscore

from ._src.client import *
from ._src.client_base import *
from ._src.enums import *
from ._src.expressions import *
from ._src.functions import *
from ._src.pyspark_client import *
from ._src.query import *

__version__ = '0.0.32'
name = "pykusto"



from datetime import timedelta, datetime

import pytest

from pykusto import Functions as f
from pykusto import column_generator as col, Query
# noinspection PyProtectedMember
from pykusto._src.expressions import _AnyTypeColumn
from test.test_base import TestBase, mock_table as t


class TestExpressions(TestBase):
    def test_contains(self):
        self.assertEqual(
            ' | where stringField contains "bar"',
            Query().where(t.stringField.contains('bar')).render(),
        )
        self.assertEqual(
            ' | where stringField contains_cs "bar"',
            Query().where(t.stringField.contains('bar', True)).render(),
        )

    def test_not_contains(self):
        self.assertEqual(
            ' | where stringField !contains "bar"',
            Query().where(t.stringField.not_contains('bar')).render(),
        )
        self.assertEqual(
            ' | where stringField !contains_cs "bar"',
            Query().where(t.stringField.not_contains('bar', True)).render(),
        )

    def test_array_access(self):
        self.assertEqual(
            ' | where (arrayField[3]) == "bar"',
            Query().where(t.arrayField[3] == 'bar').render(),
        )

    def test_array_contains(self):
        self.assertEqual(
            ' | where arrayField contains "true"',
            Query().where(t.arrayField.array_contains(True)).render(),
        )

    def test_bag_contains(self):
        self.assertEqual(
            ' | where mapField contains "2"',
            Query().where(t.mapField.bag_contains(2)).render(),
        )

    def test_not_equals(self):
        self.assertEqual(
            ' | where stringField != "bar"',
            Query().where(t.stringField != 'bar').render(),
        )

    def test_repr(self):
        self.assertEqual(
            '_StringColumn(stringField)',
            repr(t.stringField)
        )
        self.assertEqual(
            'stringField == "bar"',
            repr(t.stringField == 'bar')
        )

    def test_to_bool(self):
        self.assertEqual(
            ' | extend boolFoo = tobool(stringField)',
            Query().extend(boolFoo=t.stringField.to_bool()).render(),
        )

    def test_to_int(self):
        self.assertEqual(
            ' | extend intFoo = toint(stringField)',
            Query().extend(intFoo=t.stringField.to_int()).render(),
        )

    def test_to_long(self):
        self.assertEqual(
            ' | extend longFoo = tolong(stringField)',
            Query().extend(longFoo=t.stringField.to_long()).render(),
        )

    def test_and(self):
        self.assertEqual(
            ' | where boolField and (stringField contains "hello")',
            Query().where(t.boolField & t.stringField.contains("hello")).render(),
        )

    def test_swapped_and(self):
        self.assertEqual(
            ' | where true and boolField',
            Query().where(True & t.boolField).render(),
        )

    def test_or(self):
        self.assertEqual(
            ' | where boolField or (stringField contains "hello")',
            Query().where(t.boolField | t.stringField.contains("hello")).render(),
        )

    def test_swapped_or(self):
        self.assertEqual(
            ' | where false or boolField',
            Query().where(False | t.boolField).render(),
        )

    def test_not(self):
        self.assertEqual(
            ' | where not(stringField contains "hello")',
            Query().where(~t.stringField.contains("hello")).render(),
        )

    def test_ge(self):
        self.assertEqual(
            ' | where numField >= 10',
            Query().where(t.numField >= 10).render(),
        )

    def test_div(self):
        self.assertEqual(
            ' | extend foo = numField / 2',
            Query().extend(foo=t.numField / 2).render(),
        )

    def test_swapped_div(self):
        self.assertEqual(
            ' | extend foo = 2 / numField',
            Query().extend(foo=2 / t.numField).render(),
        )

    def test_mod(self):
        self.assertEqual(
            ' | extend foo = numField % 2',
            Query().extend(foo=t.numField % 2).render(),
        )

    def test_swapped_mod(self):
        self.assertEqual(
            ' | extend foo = 2 % numField',
            Query().extend(foo=2 % t.numField).render(),
        )

    def test_negation(self):
        self.assertEqual(
            ' | extend foo = -numField',
            Query().extend(foo=-t.numField).render(),
        )

    def test_abs(self):
        self.assertEqual(
            ' | extend foo = abs(numField)',
            Query().extend(foo=abs(t.numField)).render(),
        )

    def test_between(self):
        self.assertEqual(
            ' | where numField between (numField2 .. 100)',
            Query().where(t.numField.between(t.numField2, 100)).render(),
        )

    def test_str_equals(self):
        self.assertEqual(
            ' | where stringField =~ stringField2',
            Query().where(t.stringField.equals(t.stringField2)).render(),
        )

    def test_str_not_equals(self):
        self.assertEqual(
            ' | where stringField !~ stringField2',
            Query().where(t.stringField.not_equals(t.stringField2)).render(),
        )

    def test_str_matches(self):
        self.assertEqual(
            ' | where stringField matches regex "[a-z]+"',
            Query().where(t.stringField.matches("[a-z]+")).render(),
        )

    def test_str_starts_with(self):
        self.assertEqual(
            ' | where stringField startswith "hello"',
            Query().where(t.stringField.startswith("hello")).render(),
        )

    def test_str_ends_with(self):
        self.assertEqual(
            ' | where stringField endswith "hello"',
            Query().where(t.stringField.endswith("hello")).render(),
        )

    def test_le_date(self):
        self.assertEqual(
            'mock_table | where dateField <= datetime(2000-01-01 00:00:00.000000)',
            Query(t).where(t.dateField <= datetime(2000, 1, 1)).render(),
        )

    def test_lt_date(self):
        self.assertEqual(
            ' | where dateField < datetime(2000-01-01 00:00:00.000000)',
            Query().where(t.dateField < datetime(2000, 1, 1)).render(),
        )

    def test_ge_date(self):
        self.assertEqual(
            ' | where dateField >= datetime(2000-01-01 00:00:00.000000)',
            Query().where(t.dateField >= datetime(2000, 1, 1)).render(),
        )

    def test_gt_date(self):
        self.assertEqual(
            ' | where dateField > datetime(2000-01-01 00:00:00.000000)',
            Query().where(t.dateField > datetime(2000, 1, 1)).render(),
        )

    def test_le_timespan(self):
        self.assertEqual(
            'mock_table | where timespanField <= time(0.0:15:0.0)',
            Query(t).where(t.timespanField <= timedelta(minutes=15)).render(),
        )

    def test_lt_timespan(self):
        self.assertEqual(
            'mock_table | where timespanField < time(0.0:15:0.0)',
            Query(t).where(t.timespanField < timedelta(minutes=15)).render(),
        )

    def test_ge_timespan(self):
        self.assertEqual(
            'mock_table | where timespanField >= time(0.0:15:0.0)',
            Query(t).where(t.timespanField >= timedelta(minutes=15)).render(),
        )

    def test_gt_timespan(self):
        self.assertEqual(
            'mock_table | where timespanField > time(0.0:15:0.0)',
            Query(t).where(t.timespanField > timedelta(minutes=15)).render(),
        )

    def test_le_unknown_type(self):
        self.assertEqual(
            'mock_table | where someColumn <= 10',
            Query(t).where(col['someColumn'] <= 10).render(),
        )

    def test_lt_unknown_type(self):
        self.assertEqual(
            'mock_table | where someColumn < 10',
            Query(t).where(col['someColumn'] < 10).render(),
        )

    def test_ge_unknown_type(self):
        self.assertEqual(
            'mock_table | where someColumn >= 10',
            Query(t).where(col['someColumn'] >= 10).render(),
        )

    def test_gt_unknown_type(self):
        self.assertEqual(
            'mock_table | where someColumn > 10',
            Query(t).where(col['someColumn'] > 10).render(),
        )

    def test_add_timespan_to_date(self):
        self.assertEqual(
            ' | extend foo = dateField + time(0.1:0:0.0)',
            Query().extend(foo=t.dateField + timedelta(hours=1)).render(),
        )

    def test_add_timespan_to_timespan(self):
        self.assertEqual(
            ' | extend foo = timespanField + time(0.1:0:0.0)',
            Query().extend(foo=t.timespanField + timedelta(hours=1)).render(),
        )

    def test_add_swapped_timespan_to_timespan(self):
        self.assertEqual(
            ' | extend foo = time(0.1:0:0.0) + timespanField',
            Query().extend(foo=timedelta(hours=1) + t.timespanField).render(),
        )

    def test_subtract_timespan_from_timespan(self):
        self.assertEqual(
            ' | extend foo = timespanField - time(0.1:0:0.0)',
            Query().extend(foo=t.timespanField - timedelta(hours=1)).render(),
        )

    def test_swapped_subtract_timespan_from_timespan(self):
        self.assertEqual(
            ' | extend foo = time(0.1:0:0.0) - timespanField',
            Query().extend(foo=timedelta(hours=1) - t.timespanField).render(),
        )

    def test_sub_timespan(self):
        self.assertEqual(
            ' | extend foo = dateField - time(0.1:0:0.0)',
            Query().extend(foo=t.dateField - timedelta(hours=1)).render(),
        )

    def test_sub_datetime(self):
        self.assertEqual(
            ' | extend foo = dateField - datetime(2020-01-01 00:00:00.000000)',
            Query().extend(foo=t.dateField - datetime(2020, 1, 1)).render(),
        )

    def test_sub_from_datetime(self):
        self.assertEqual(
            ' | extend foo = datetime(2020-01-01 00:00:00.000000) - dateField',
            Query().extend(foo=datetime(2020, 1, 1) - t.dateField).render(),
        )

    def test_sub_from_number(self):
        self.assertEqual(
            ' | extend foo = 3 - numField',
            Query().extend(foo=3 - t.numField).render(),
        )

    def test_sub_date_unknown_type(self):
        self.assertEqual(
            ' | extend foo = dateField - (case(boolField, bar, baz))',
            Query().extend(foo=t.dateField - f.case(t.boolField, col.bar, col.baz)).render(),
        )

    def test_sub_date_unknown_column(self):
        self.assertEqual(
            ' | extend foo = dateField - bar',
            Query().extend(foo=t.dateField - col.bar).render(),
        )

    def test_sub_unknown_type_number(self):
        self.assertEqual(
            ' | extend foo = cos(bar - numField)',
            Query().extend(foo=(col.bar - t.numField).cos()).render(),
        )

    def test_sub_unknown_type_datetime(self):
        self.assertEqual(
            ' | extend foo = ago(bar - dateField)',
            Query().extend(foo=(col.bar - t.dateField).ago()).render(),
        )

    def test_sub_unknown_type_timespan(self):
        self.assertEqual(
            ' | extend foo = bar - timespanField',
            Query().extend(foo=col.bar - t.timespanField).render(),
        )

    def test_bin_auto(self):
        self.assertEqual(
            ' | extend foo = bin_auto(dateField)',
            Query().extend(foo=t.dateField.bin_auto()).render(),
        )

    def test_array_access_expression_index(self):
        self.assertEqual(
            ' | where (arrayField[numField * 2]) == "bar"',
            Query().where(t.arrayField[t.numField * 2] == 'bar').render(),
        )

    def test_array_access_yields_any_expression(self):
        self.assertEqual(
            ' | where (cos(arrayField[3])) < 1',
            Query().where(t.arrayField[3].cos() < 1).render(),
        )

    def test_mapping_access(self):
        self.assertEqual(
            ' | where (mapField["key"]) == "bar"',
            Query().where(t.mapField['key'] == 'bar').render(),
        )

    def test_mapping_access_attribute(self):
        self.assertEqual(
            ' | where (mapField.key) == "bar"',
            Query().where(t.mapField.key == 'bar').render(),
        )

    def test_mapping_access_expression_index(self):
        self.assertEqual(
            ' | where (mapField[stringField]) == "bar"',
            Query().where(t.mapField[t.stringField] == 'bar').render(),
        )

    def test_mapping_access_yields_any_expression(self):
        self.assertEqual(
            ' | where (mapField["key"]) contains "substr"',
            Query().where(t.mapField['key'].contains("substr")).render(),
        )

    def test_dynamic(self):
        self.assertEqual(
            ' | where (mapField["foo"][0].bar[1][2][tolower(stringField)]) > time(1.0:0:0.0)',
            Query().where(t.mapField['foo'][0].bar[1][2][t.stringField.lower()] > timedelta(1)).render(),
        )

    def test_assign_to(self):
        self.assertEqual(
            " | extend numFieldNew = numField * 2",
            Query().extend((t.numField * 2).assign_to(col.numFieldNew)).render(),
        )
        self.assertEqual(
            " | extend foo = numField * 2",
            Query().extend(foo=(t.numField * 2)).render(),
        )

    def test_extend_const(self):
        self.assertEqual(
            ' | extend foo = 5, bar = "bar", other_col = stringField',
            Query().extend(foo=5, bar="bar", other_col=t.stringField).render(),
        )

    def test_between_date(self):
        self.assertEqual(
            " | where dateField between (datetime(2020-01-01 00:00:00.000000) .. datetime(2020-01-31 00:00:00.000000))",
            Query().where(t.dateField.between(datetime(2020, 1, 1), datetime(2020, 1, 31))).render(),
        )

    def test_between_timespan(self):
        self.assertEqual(
            " | where timespanField between (time(0.0:0:0.0) .. time(0.3:0:0.0))",
            Query().where(t.timespanField.between(timedelta(0), timedelta(hours=3))).render(),
        )

    def test_is_empty(self):
        self.assertEqual(
            'isempty(stringField)',
            t.stringField.is_empty().kql,
        )

    def test_column_with_dot(self):
        self.assertEqual(
            " | project ['foo.bar']",
            Query().project(t['foo.bar']).render(),
        )

    def test_is_in(self):
        self.assertEqual(
            ' | where stringField in ("A", "B", "C")',
            Query().where(t.stringField.is_in(["A", "B", "C"], True)).render()
        )
        self.assertEqual(
            ' | where stringField in~ ("[", "[[", "]")',
            Query().where(t.stringField.is_in(['[', "[[", "]"])).render()
        )
        self.assertRaises(
            NotImplementedError("'in' not supported. Instead use '.is_in()'"),
            lambda: t.stringField in t.stringField2
        )

    def test_not_in(self):
        self.assertEqual(
            ' | where stringField !in ("A", "B", "C")',
            Query().where(t.stringField.not_in(["A", "B", "C"], True)).render()
        )
        self.assertEqual(
            ' | where stringField !in~ ("[", "[[", "]")',
            Query().where(t.stringField.not_in(['[', "[[", "]"])).render()
        )

    def test_is_in_expression(self):
        self.assertEqual(
            ' | where set_has_element(arrayField, stringField)',
            Query().where(t.stringField.is_in(t.arrayField, True)).render()
        )

    def test_not_in_expression(self):
        self.assertEqual(
            ' | where arrayField !contains stringField',
            Query().where(t.stringField.not_in(t.arrayField, False)).render()
        )

    def test_not_in_cs_expression(self):
        self.assertEqual(
            ' | where arrayField !contains_cs stringField',
            Query().where(t.stringField.not_in(t.arrayField, True)).render()
        )

    def test_has(self):
        self.assertEqual(
            ' | where stringField has "test"',
            Query().where(t.stringField.has("test")).render()
        )

    def test_has_not(self):
        self.assertEqual(
            ' | where stringField !has "test"',
            Query().where(t.stringField.has_not("test")).render()
        )

    def test_has_not_cs(self):
        self.assertEqual(
            ' | where stringField !has_cs "test"',
            Query().where(t.stringField.has_not("test", True)).render()
        )

    def test_has_cs(self):
        self.assertEqual(
            ' | where stringField has_cs "test"',
            Query().where(t.stringField.has("test", case_sensitive=True)).render()
        )

    def test_has_any(self):
        self.assertEqual(
            ' | where stringField has_any ("field", "string")',
            Query().where(t.stringField.has_any(["field", "string"])).render()
        )

    @pytest.mark.skip(reason="Re-enable once this is resoled: https://github.com/agronholm/typeguard/issues/159")
    def test_has_any_bad_argument(self):
        self.assertRaises(
            AssertionError("Compared array must be a list of tabular, scalar, or literal expressions"),
            lambda: t.stringField.has_any(t.stringField2)
        )

    def test_column_generator(self):
        field1 = col.foo
        field2 = col['foo.bar']
        self.assertIsInstance(field1, _AnyTypeColumn)
        self.assertIsInstance(field2, _AnyTypeColumn)
        self.assertEqual('foo', field1.get_name())
        self.assertEqual('foo.bar', field2.get_name())

    def test_column_name_quoting(self):
        self.assertEqual(
            ' | where [\'title\'] has "test"',
            Query().where(t.title.has("test")).render()
        )
        self.assertEqual(
            ' | where [\'stringField\'] has "test"',
            Query().where(col.of('stringField').has("test")).render()
        )

    def test_multiply_number_column(self):
        self.assertEqual(
            ' | where (todouble(100 * numberField)) > 0.2',
            Query().where(f.to_double(100 * t.numberField) > 0.2).render(),
        )

    def test_add_number_column(self):
        self.assertEqual(
            ' | where (todouble(100 + numberField)) > 0.2',
            Query().where(f.to_double(100 + t.numberField) > 0.2).render(),
        )

    def test_multiply_number_expression(self):
        self.assertEqual(
            ' | where (100 * (todouble(numberField))) > 0.2',
            Query().where(100 * f.to_double(t.numberField) > 0.2).render(),
        )

    def test_column_with_digits(self):
        self.assertEqual(
            " | where (['100'] * (todouble(numberField))) > 0.2",
            Query().where(col['100'] * f.to_double(t.numberField) > 0.2).render(),
        )

    def test_boolean_operators(self):
        self.assertRaises(
            TypeError(
                "Conversion of expression to boolean is not allowed, to prevent accidental use of the logical operators: 'and', 'or', and 'not'. "
                "Instead either use the bitwise operators '&', '|' and '~' (but note the difference in operator precedence!), "
                "or the functions 'all_of', 'any_of' and 'not_of'"
            ),
            lambda: (t.boolField and t.numField > 10)
        )



from datetime import datetime, timedelta
from typing import Any, List, Tuple, Mapping, Optional, Iterable
from typing import Union

from .keywords import _KUSTO_KEYWORDS
from .kql_converters import KQL
from .type_utils import _plain_expression, _aggregation_expression, PythonTypes, _kql_converter, _KustoType, \
    _typed_column, _TypeRegistrar, _get_base_types, _NUMBER_TYPES

ExpressionType = Union[PythonTypes, 'BaseExpression']
StringType = Union[str, '_StringExpression']
BooleanType = Union[bool, '_BooleanExpression']
NumberType = Union[int, float, '_NumberExpression']
ArrayType = Union[List, Tuple, '_ArrayExpression']
MappingType = Union[Mapping, '_MappingExpression']
DatetimeType = Union[datetime, '_DatetimeExpression']
TimespanType = Union[timedelta, '_TimespanExpression']
DynamicType = Union[ArrayType, MappingType]
ComparableType = Union[DatetimeType, TimespanType, NumberType]
OrderedType = Union[ComparableType, StringType]


# All classes in the same file to prevent circular dependencies

class BaseExpression:
    kql: KQL

    # We would prefer to use 'abc' to make the class abstract, but this can be done only if there is at least one
    # abstract method, which we don't have here. Overriding __new___ is the next best solution.
    def __new__(cls, *args, **kwargs) -> 'BaseExpression':
        assert cls is not BaseExpression, "BaseExpression is abstract"
        return object.__new__(cls)

    def __init__(self, kql: Union[KQL, 'BaseExpression']) -> None:
        if isinstance(kql, BaseExpression):
            self.kql = kql.kql
            return
        assert isinstance(kql, str), "Either expression or KQL required"
        self.kql = kql

    def __repr__(self) -> str:
        return str(self.kql)

    def __bool__(self) -> bool:
        raise TypeError(
            "Conversion of expression to boolean is not allowed, to prevent accidental use of the logical operators: 'and', 'or', and 'not'. "
            "Instead either use the bitwise operators '&', '|' and '~' (but note the difference in operator precedence!), "
            "or the functions 'all_of', 'any_of' and 'not_of'"
        )

    def as_subexpression(self) -> KQL:
        return KQL(f'({self.kql})')

    def get_type(self) -> '_StringExpression':
        return _StringExpression(KQL(f'gettype({self.kql})'))

    def __hash__(self, mod: Union['_NumberExpression', int] = None) -> '_StringExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/hashfunction
        """
        if mod is not None:
            return _StringExpression(KQL(f'hash({self.kql}, {_to_kql(mod)})'))
        return _StringExpression(KQL(f'hash({self.kql})'))

    def hash_sha256(self) -> '_StringExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/sha256hashfunction
        """
        return _StringExpression(KQL(f'hash_sha256({self.kql})'))

    def is_empty(self) -> '_BooleanExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/isemptyfunction
        """
        return _BooleanExpression(KQL(f'isempty({self.kql})'))

    def is_not_empty(self) -> '_BooleanExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/isnotemptyfunction
        """
        return _BooleanExpression(KQL(f'isnotempty({self.kql})'))

    @staticmethod
    def base_binary_op(
            left: ExpressionType, operator: str, right: ExpressionType, result_type: Optional[_KustoType]
    ) -> 'BaseExpression':
        registrar = _plain_expression
        fallback = AnyExpression
        if isinstance(left, AggregationExpression) or isinstance(right, AggregationExpression):
            registrar = _aggregation_expression
            fallback = _AnyAggregationExpression
        return_type = fallback if result_type is None else registrar.registry[result_type]
        return return_type(KQL(f'{_to_kql(left, True)}{operator}{_to_kql(right, True)}'))

    def __eq__(self, other: ExpressionType) -> '_BooleanExpression':
        return _BooleanExpression.binary_op(self, ' == ', other)

    def __ne__(self, other: ExpressionType) -> '_BooleanExpression':
        return _BooleanExpression.binary_op(self, ' != ', other)

    def is_in(self, other: DynamicType, case_sensitive: bool = False) -> '_BooleanExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/scalar-data-types/dynamic#operators-and-functions-over-dynamic-types
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/inoperator
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/datatypes-string-operators
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/sethaselementfunction

        Best practices (`full list <https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/best-practices>`_):

        * Use `case_sensitive=True` when possible
        """
        if isinstance(other, (List, Tuple)):
            # For a literal array, we can use 'in'
            # The following RHS is the only place where a literal list does not require being surrounded by 'dynamic()'
            return _BooleanExpression(KQL(f'{self.kql} {"in" if case_sensitive else "in~"} ({", ".join(map(_to_kql, other))})'))
        # Otherwise, Kusto does not accept 'in', and we need to use 'set_has_element', which is always case-sensitive
        assert case_sensitive, "Member test for non-literal array cannot be case insensitive"
        return _BooleanExpression(KQL(f'set_has_element({_to_kql(other)}, {self.kql})'))

    def not_in(self, other: DynamicType, case_sensitive: bool = False) -> '_BooleanExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/scalar-data-types/dynamic#operators-and-functions-over-dynamic-types
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/inoperator
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/datatypes-string-operators

        Best practices (`full list <https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/best-practices>`_):

        * Use `case_sensitive=True` when possible
        """
        if isinstance(other, (List, Tuple)):
            # For a literal array, we can use 'in'
            # The following RHS is the only place where a literal list does not require being surrounded by 'dynamic()'
            return _BooleanExpression(KQL(f'{self.kql} {"!in" if case_sensitive else "!in~"} ({", ".join(map(_to_kql, other))})'))
        # Otherwise, for some reason Kusto does not accept 'in', and we need to use 'contains' as if 'other' was a string
        return _BooleanExpression.binary_op(other, ' !contains_cs ' if case_sensitive else ' !contains ', self)

    def is_null(self) -> '_BooleanExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/isnullfunction
        """
        return _BooleanExpression(KQL(f'isnull({self.kql})'))

    def is_not_null(self) -> '_BooleanExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/isnotnullfunction
        """
        return _BooleanExpression(KQL(f'isnotnull({self.kql})'))

    def __contains__(self, other: Any) -> bool:
        """
        Deliberately not implemented, because "not in" inverses the result of this method, and there is no way to
        override it
        """
        raise NotImplementedError("'in' not supported. Instead use '.is_in()'")

    def to_bool(self) -> '_BooleanExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/toboolfunction
        """
        return _BooleanExpression(KQL(f'tobool({self.kql})'))

    def to_string(self) -> '_StringExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/tostringfunction
        """
        return _StringExpression(KQL(f'tostring({self.kql})'))

    def to_int(self) -> '_NumberExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/tointfunction
        """
        return _NumberExpression(KQL(f'toint({self.kql})'))

    def to_long(self) -> '_NumberExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/tolongfunction
        """
        return _NumberExpression(KQL(f'tolong({self.kql})'))

    def assign_to_single_column(self, column: 'BaseColumn') -> '_AssignmentToSingleColumn':
        return _AssignmentToSingleColumn(column, self)

    def assign_to_multiple_columns(self, *columns: '_AnyTypeColumn') -> '_AssignmentBase':
        """
        This method exists for the sole purpose of providing an informative error message.
        """
        raise ValueError("Only arrays can be assigned to multiple columns")

    def assign_to(self, *columns: 'BaseColumn') -> '_AssignmentBase':
        if len(columns) == 0:
            # Unspecified column name
            return _AssignmentBase(None, self)
        if len(columns) == 1:
            return self.assign_to_single_column(columns[0])
        return self.assign_to_multiple_columns(*columns)


@_plain_expression(_KustoType.BOOL)
class _BooleanExpression(BaseExpression):
    @staticmethod
    def binary_op(left: ExpressionType, operator: str, right: ExpressionType) -> '_BooleanExpression':
        # noinspection PyTypeChecker
        return BaseExpression.base_binary_op(left, operator, right, _KustoType.BOOL)

    def __and__(self, other: BooleanType) -> '_BooleanExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/logicaloperators
        """
        return _BooleanExpression.binary_op(self, ' and ', other)

    def __rand__(self, other: BooleanType) -> '_BooleanExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/logicaloperators
        """
        return _BooleanExpression.binary_op(other, ' and ', self)

    def __or__(self, other: BooleanType) -> '_BooleanExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/logicaloperators
        """
        return _BooleanExpression.binary_op(self, ' or ', other)

    def __ror__(self, other: BooleanType) -> '_BooleanExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/logicaloperators
        """
        return _BooleanExpression.binary_op(other, ' or ', self)

    def __invert__(self) -> '_BooleanExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/notfunction
        Note that using the Python 'not' does not have the desired effect, because unfortunately its behavior cannot be overridden.
        """
        return _BooleanExpression(KQL(f'not({self.kql})'))


@_plain_expression(*_NUMBER_TYPES)
class _NumberExpression(BaseExpression):
    @staticmethod
    def binary_op(left: NumberType, operator: str, right: NumberType) -> '_NumberExpression':
        # noinspection PyTypeChecker
        return BaseExpression.base_binary_op(left, operator, right, _KustoType.INT)

    def __lt__(self, other: NumberType) -> _BooleanExpression:
        return _BooleanExpression.binary_op(self, ' < ', other)

    def __le__(self, other: NumberType) -> _BooleanExpression:
        return _BooleanExpression.binary_op(self, ' <= ', other)

    def __gt__(self, other: NumberType) -> _BooleanExpression:
        return _BooleanExpression.binary_op(self, ' > ', other)

    def __ge__(self, other: NumberType) -> _BooleanExpression:
        return _BooleanExpression.binary_op(self, ' >= ', other)

    def __add__(self, other: NumberType) -> '_NumberExpression':
        return _NumberExpression.binary_op(self, ' + ', other)

    def __radd__(self, other: NumberType) -> '_NumberExpression':
        return _NumberExpression.binary_op(other, ' + ', self)

    def __sub__(self, other: NumberType) -> '_NumberExpression':
        return _NumberExpression.binary_op(self, ' - ', other)

    def __rsub__(self, other: NumberType) -> '_NumberExpression':
        return _NumberExpression.binary_op(other, ' - ', self)

    def __mul__(self, other: NumberType) -> '_NumberExpression':
        return _NumberExpression.binary_op(self, ' * ', other)

    def __rmul__(self, other: NumberType) -> '_NumberExpression':
        return _NumberExpression.binary_op(other, ' * ', self)

    def __truediv__(self, other: NumberType) -> '_NumberExpression':
        return _NumberExpression.binary_op(self, ' / ', other)

    def __rtruediv__(self, other: NumberType) -> '_NumberExpression':
        return _NumberExpression.binary_op(other, ' / ', self)

    def __mod__(self, other: NumberType) -> '_NumberExpression':
        return _NumberExpression.binary_op(self, ' % ', other)

    def __rmod__(self, other: NumberType) -> '_NumberExpression':
        return _NumberExpression.binary_op(other, ' % ', self)

    def __neg__(self) -> '_NumberExpression':
        return _NumberExpression(KQL(f'-{self.kql}'))

    def __abs__(self) -> '_NumberExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/abs-function
        """
        return _NumberExpression(KQL(f'abs({self.kql})'))

    def between(self, lower: NumberType, upper: NumberType) -> _BooleanExpression:
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/betweenoperator
        """
        return _BooleanExpression(KQL(f'{self.kql} between ({_to_kql(lower, True)} .. {_to_kql(upper, True)})'))

    def acos(self) -> '_NumberExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/acosfunction
        """
        return _NumberExpression(KQL(f'acos({self.kql})'))

    def cos(self) -> '_NumberExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/cosfunction
        """
        return _NumberExpression(KQL(f'cos({self.kql})'))

    def floor(self, round_to: NumberType) -> '_NumberExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/floorfunction
        """
        return _NumberExpression(KQL(f'floor({self.kql}, {_to_kql(round_to)})'))

    def bin(self, round_to: NumberType) -> 'BaseExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/binfunction
        """
        return _NumberExpression(KQL(f'bin({self.kql}, {_to_kql(round_to)})'))

    def bin_at(self, round_to: NumberType, fixed_point: NumberType) -> 'BaseExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/binatfunction
        """
        return _NumberExpression(KQL(f'bin_at({self.kql}, {_to_kql(round_to)}, {_to_kql(fixed_point)})'))

    def bin_auto(self) -> 'BaseExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/bin-autofunction
        """
        return _NumberExpression(KQL(f'bin_auto({self.kql})'))

    def ceiling(self) -> '_NumberExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/ceilingfunction
        """
        return _NumberExpression(KQL(f'ceiling({self.kql})'))

    def exp(self) -> '_NumberExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/exp-function
        """
        return _NumberExpression(KQL(f'exp({self.kql})'))

    def exp10(self) -> '_NumberExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/exp10-function
        """
        return _NumberExpression(KQL(f'exp10({self.kql})'))

    def exp2(self) -> '_NumberExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/exp2-function
        """
        return _NumberExpression(KQL(f'exp2({self.kql})'))

    def isfinite(self) -> _BooleanExpression:
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/isfinitefunction
        """
        return _BooleanExpression(KQL(f'isfinite({self.kql})'))

    def is_inf(self) -> _BooleanExpression:
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/isinffunction
        """
        return _BooleanExpression(KQL(f'isinf({self.kql})'))

    def is_nan(self) -> _BooleanExpression:
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/isnanfunction
        """
        return _BooleanExpression(KQL(f'isnan({self.kql})'))

    def log(self) -> '_NumberExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/log-function
        """
        return _NumberExpression(KQL(f'log({self.kql})'))

    def log10(self) -> '_NumberExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/log10-function
        """
        return _NumberExpression(KQL(f'log10({self.kql})'))

    def log2(self) -> '_NumberExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/log2-function
        """
        return _NumberExpression(KQL(f'log2({self.kql})'))

    def log_gamma(self) -> '_NumberExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/loggammafunction
        """
        return _NumberExpression(KQL(f'loggamma({self.kql})'))

    def round(self, precision: NumberType = None) -> '_NumberExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/roundfunction
        """
        return _NumberExpression(KQL(f'round({self.kql})' if precision is None else f'round({self.kql}, {_to_kql(precision)})'))


@_plain_expression(_KustoType.STRING)
class _StringExpression(BaseExpression):
    # We would like to allow using len(), but Python requires it to return an int, so we can't
    def string_size(self) -> _NumberExpression:
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/stringsizefunction
        """
        return _NumberExpression(KQL(f'string_size({self.kql})'))

    def split(self, delimiter: StringType, requested_index: NumberType = None) -> '_ArrayExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/splitfunction
        """
        if requested_index is None:
            return _ArrayExpression(KQL(f'split({self.kql}, {_to_kql(delimiter)})'))
        return _ArrayExpression(KQL(f'split({self.kql}, {_to_kql(delimiter)}, {_to_kql(requested_index)})'))

    def equals(self, other: StringType, case_sensitive: bool = False) -> _BooleanExpression:
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/datatypes-string-operators

        Best practices (`full list <https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/best-practices>`_):

        * Use `case_sensitive=True` when possible
        """
        return _BooleanExpression.binary_op(self, ' == ' if case_sensitive else ' =~ ', other)

    def not_equals(self, other: StringType, case_sensitive: bool = False) -> _BooleanExpression:
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/datatypes-string-operators

        Best practices (`full list <https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/best-practices>`_):

        * Use `case_sensitive=True` when possible
        """
        return _BooleanExpression.binary_op(self, ' != ' if case_sensitive else ' !~ ', other)

    def matches(self, regex: StringType) -> '_BooleanExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/datatypes-string-operators
        """
        return _BooleanExpression.binary_op(self, ' matches regex ', regex)

    def contains(self, other: StringType, case_sensitive: bool = False) -> _BooleanExpression:
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/datatypes-string-operators

        Best practices (`full list <https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/best-practices>`_):

        * When looking for full tokens, `has` works better, since it doesn't look for substrings.
        * Use `case_sensitive=True` when possible
        """
        return _BooleanExpression.binary_op(self, ' contains_cs ' if case_sensitive else ' contains ', other)

    def not_contains(self, other: StringType, case_sensitive: bool = False) -> _BooleanExpression:
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/datatypes-string-operators

        Best practices (`full list <https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/best-practices>`_):

        * When looking for full tokens, `not_has` works better, since it doesn't look for substrings.
        * Use `case_sensitive=True` when possible
        """
        return _BooleanExpression.binary_op(self, ' !contains_cs ' if case_sensitive else ' !contains ', other)

    def startswith(self, other: StringType, case_sensitive: bool = False) -> _BooleanExpression:
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/datatypes-string-operators
        """
        return _BooleanExpression.binary_op(self, ' startswith_cs ' if case_sensitive else ' startswith ', other)

    def endswith(self, other: StringType, case_sensitive: bool = False) -> _BooleanExpression:
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/datatypes-string-operators
        """
        return _BooleanExpression.binary_op(self, ' endswith_cs ' if case_sensitive else ' endswith ', other)

    def lower(self) -> '_StringExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/tolowerfunction

        Don't use this method for case-insensitive comparisons. Instead use the `equals` method with `case_sensitive=False`
        (see `best practices <https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/best-practices>`_).
        """
        return _StringExpression(KQL(f'tolower({self.kql})'))

    def upper(self) -> '_StringExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/toupperfunction

        Don't use this method for case-insensitive comparisons. Instead use the `equals` method with `case_sensitive=False`
        (see `best practices <https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/best-practices>`_).
        """
        return _StringExpression(KQL(f'toupper({self.kql})'))

    def is_utf8(self) -> _BooleanExpression:
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/isutf8
        """
        return _BooleanExpression(KQL(f'isutf8({self.kql})'))

    def has(self, exp: StringType, case_sensitive: bool = False) -> '_BooleanExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/datatypes-string-operators
        """
        return _BooleanExpression(KQL(
            f'{self.as_subexpression()} {"has_cs" if case_sensitive else "has"} {_to_kql(exp, True)}'
        ))

    def has_not(self, exp: StringType, case_sensitive: bool = False) -> '_BooleanExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/datatypes-string-operators
        """
        return _BooleanExpression(KQL(
            f'{self.as_subexpression()} {"!has_cs" if case_sensitive else "!has"} {_to_kql(exp, True)}'
        ))

    def has_any(self, other: Union[List, Tuple]) -> '_BooleanExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/has-anyoperator
        Please notice that this implementation does not support tabular expression inputs currently
        """
        assert isinstance(other, (List, Tuple)), "Compared array must be a list of tabular, scalar, or literal expressions"
        return _BooleanExpression(KQL(f'{self.kql} has_any ({", ".join(map(_to_kql, other))})'))


@_plain_expression(_KustoType.DATETIME)
class _DatetimeExpression(BaseExpression):
    @staticmethod
    def binary_op(left: ExpressionType, operator: str, right: ExpressionType) -> '_DatetimeExpression':
        # noinspection PyTypeChecker
        return BaseExpression.base_binary_op(left, operator, right, _KustoType.DATETIME)

    def __lt__(self, other: DatetimeType) -> _BooleanExpression:
        return _BooleanExpression.binary_op(self, ' < ', other)

    def __le__(self, other: DatetimeType) -> _BooleanExpression:
        return _BooleanExpression.binary_op(self, ' <= ', other)

    def __gt__(self, other: DatetimeType) -> _BooleanExpression:
        return _BooleanExpression.binary_op(self, ' > ', other)

    def __ge__(self, other: DatetimeType) -> _BooleanExpression:
        return _BooleanExpression.binary_op(self, ' >= ', other)

    def __add__(self, other: TimespanType) -> '_DatetimeExpression':
        # https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/datetime-timespan-arithmetic
        return _DatetimeExpression.binary_op(self, ' + ', other)

    def __sub__(self, other: Union[DatetimeType, TimespanType]) -> Union['_DatetimeExpression', '_TimespanExpression']:
        # https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/datetime-timespan-arithmetic

        # noinspection PyTypeChecker
        base_types = _get_base_types(other)
        possible_types = base_types & {_KustoType.DATETIME, _KustoType.TIMESPAN}
        assert len(possible_types) > 0, "Invalid type subtracted from datetime"
        if len(possible_types) == 2:
            return_type = AnyExpression
        else:
            return_type = _TimespanExpression if next(iter(possible_types)) == _KustoType.DATETIME else _DatetimeExpression
        return return_type(_DatetimeExpression.binary_op(self, ' - ', other))

    def __rsub__(self, other: DatetimeType) -> '_TimespanExpression':
        # https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/datetime-timespan-arithmetic
        return _TimespanExpression.binary_op(other, ' - ', self)

    def between(self, lower: DatetimeType, upper: DatetimeType) -> _BooleanExpression:
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/betweenoperator
        """
        return _BooleanExpression(KQL(f'{self.kql} between ({_to_kql(lower, True)} .. {_to_kql(upper, True)})'))

    def floor(self, round_to: TimespanType) -> '_DatetimeExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/floorfunction
        """
        return _DatetimeExpression(KQL(f'floor({self.kql}, {_to_kql(round_to)})'))

    def bin(self, round_to: TimespanType) -> 'BaseExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/binfunction
        """
        return _DatetimeExpression(KQL(f'bin({self.kql}, {_to_kql(round_to)})'))

    def bin_at(self, round_to: TimespanType, fixed_point: DatetimeType) -> 'BaseExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/binatfunction
        """
        return _DatetimeExpression(KQL(f'bin_at({self.kql}, {_to_kql(round_to)}, {_to_kql(fixed_point)})'))

    def bin_auto(self) -> '_DatetimeExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/bin-autofunction
        """
        return _DatetimeExpression(KQL(f'bin_auto({self.kql})'))

    def end_of_day(self, offset: NumberType = None) -> '_DatetimeExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/endofdayfunction
        """
        if offset is None:
            res = f'endofday({self.kql})'
        else:
            res = f'endofday({self.kql}, {_to_kql(offset)})'
        return _DatetimeExpression(KQL(res))

    def end_of_month(self, offset: NumberType = None) -> '_DatetimeExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/endofmonthfunction
        """
        if offset is None:
            res = f'endofmonth({self.kql})'
        else:
            res = f'endofmonth({self.kql}, {_to_kql(offset)})'
        return _DatetimeExpression(KQL(res))

    def end_of_week(self, offset: NumberType = None) -> '_DatetimeExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/endofweekfunction
        """
        if offset is None:
            res = f'endofweek({self.kql})'
        else:
            res = f'endofweek({self.kql}, {_to_kql(offset)})'
        return _DatetimeExpression(KQL(res))

    def end_of_year(self, offset: NumberType = None) -> '_DatetimeExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/endofyearfunction
        """
        if offset is None:
            res = f'endofyear({self.kql})'
        else:
            res = f'endofyear({self.kql}, {_to_kql(offset)})'
        return _DatetimeExpression(KQL(res))

    def format_datetime(self, format_string: StringType) -> _StringExpression:
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/format-datetimefunction
        """
        return _StringExpression(KQL(f'format_datetime({self.kql}, {_to_kql(format_string)})'))

    def get_month(self) -> _NumberExpression:
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/getmonthfunction
        """
        return _NumberExpression(KQL(f'getmonth({self.kql})'))

    def get_year(self) -> _NumberExpression:
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/getyearfunction
        """
        return _NumberExpression(KQL(f'getyear({self.kql})'))

    def hour_of_day(self) -> _NumberExpression:
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/hourofdayfunction
        """
        return _NumberExpression(KQL(f'hourofday({self.kql})'))

    def start_of_day(self, offset: NumberType = None) -> '_DatetimeExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/startofdayfunction
        """
        return _DatetimeExpression(KQL(f'startofday({self.kql})' if offset is None else f'startofday({self.kql}, {_to_kql(offset)})'))

    def start_of_month(self, offset: NumberType = None) -> '_DatetimeExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/startofmonthfunction
        """
        return _DatetimeExpression(KQL(f'startofmonth({self.kql})' if offset is None else f'startofmonth({self.kql}, {_to_kql(offset)})'))

    def start_of_week(self, offset: NumberType = None) -> '_DatetimeExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/startofweekfunction
        """
        return _DatetimeExpression(KQL(f'startofweek({self.kql})' if offset is None else f'startofweek({self.kql}, {_to_kql(offset)})'))

    def start_of_year(self, offset: NumberType = None) -> '_DatetimeExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/startofyearfunction
        """
        return _DatetimeExpression(KQL(f'startofyear({self.kql})' if offset is None else 'startofyear({self.kql}, {to_kql(offset)})'))

    def day_of_week(self) -> '_TimespanExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/dayofweekfunction
        """
        return _TimespanExpression(KQL(f'dayofweek({self.kql})'))


@_plain_expression(_KustoType.TIMESPAN)
class _TimespanExpression(BaseExpression):
    @staticmethod
    def binary_op(left: ExpressionType, operator: str, right: ExpressionType) -> '_TimespanExpression':
        # noinspection PyTypeChecker
        return BaseExpression.base_binary_op(left, operator, right, _KustoType.TIMESPAN)

    def __add__(self, other: TimespanType) -> '_TimespanExpression':
        # https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/datetime-timespan-arithmetic
        return _TimespanExpression.binary_op(self, ' + ', other)

    def __radd__(self, other: TimespanType) -> '_TimespanExpression':
        # https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/datetime-timespan-arithmetic
        return _TimespanExpression.binary_op(other, ' + ', self)

    def __sub__(self, other: TimespanType) -> '_TimespanExpression':
        # https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/datetime-timespan-arithmetic
        return _TimespanExpression.binary_op(self, ' - ', other)

    def __rsub__(self, other: TimespanType) -> '_TimespanExpression':
        # https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/datetime-timespan-arithmetic
        return _TimespanExpression.binary_op(other, ' - ', self)

    def __lt__(self, other: TimespanType) -> _BooleanExpression:
        return _BooleanExpression.binary_op(self, ' < ', other)

    def __le__(self, other: TimespanType) -> _BooleanExpression:
        return _BooleanExpression.binary_op(self, ' <= ', other)

    def __gt__(self, other: TimespanType) -> _BooleanExpression:
        return _BooleanExpression.binary_op(self, ' > ', other)

    def __ge__(self, other: TimespanType) -> _BooleanExpression:
        return _BooleanExpression.binary_op(self, ' >= ', other)

    def ago(self) -> _DatetimeExpression:
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/agofunction
        """
        return _DatetimeExpression(KQL(f'ago({_to_kql(self)})'))

    def bin(self, round_to: TimespanType) -> 'BaseExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/binfunction
        """
        return _TimespanExpression(KQL(f'bin({self.kql}, {_to_kql(round_to)})'))

    def bin_at(self, round_to: TimespanType, fixed_point: TimespanType) -> 'BaseExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/binatfunction
        """
        return _TimespanExpression(KQL(f'bin_at({self.kql}, {_to_kql(round_to)}, {_to_kql(fixed_point)})'))

    def bin_auto(self) -> 'BaseExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/bin-autofunction
        """
        return _TimespanExpression(KQL(f'bin_auto({self.kql})'))

    def format_timespan(self, format_string: StringType) -> _StringExpression:
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/format-timespanfunction
        """
        return _StringExpression(KQL(f'format_timespan({self.kql}, {_to_kql(format_string)})'))

    def between(self, lower: TimespanType, upper: TimespanType) -> _BooleanExpression:
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/betweenoperator
        """
        return _BooleanExpression(KQL(f'{self.kql} between ({_to_kql(lower, True)} .. {_to_kql(upper, True)})'))


class _BaseDynamicExpression(BaseExpression):
    # We would prefer to use 'abc' to make the class abstract, but this can be done only if there is at least one
    # abstract method, which we don't have here. Overriding __new___ is the next best solution.
    def __new__(cls, *args, **kwargs) -> '_BaseDynamicExpression':
        assert cls is not _BaseDynamicExpression, "BaseDynamicExpression is abstract"
        return object.__new__(cls)

    def __getitem__(self, index: Union[StringType, NumberType]) -> 'AnyExpression':
        return AnyExpression(KQL(f'{self.kql}[{_to_kql(index)}]'))

    def contains(self, other: ExpressionType) -> '_BooleanExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/scalar-data-types/dynamic#operators-and-functions-over-dynamic-types
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/inoperator
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/datatypes-string-operators
        """
        # For some reason Kusto does not accept 'in', and we need to use 'contains' as if this were a string
        if not isinstance(other, (BaseExpression, str)):
            # When 'other' is a literal, it has to be a string, because syntactically 'contains' works only on strings.
            other = str(_to_kql(other))
        return _BooleanExpression.binary_op(self, ' contains ', other)


@_plain_expression(_KustoType.ARRAY)
class _ArrayExpression(_BaseDynamicExpression):
    def __getitem__(self, index: NumberType) -> 'AnyExpression':
        return super().__getitem__(index)

    # We would like to allow using len(), but Python requires it to return an int, so we can't
    def array_length(self) -> _NumberExpression:
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/arraylengthfunction
        """
        return _NumberExpression(KQL(f'array_length({self.kql})'))

    def array_contains(self, other: ExpressionType) -> '_BooleanExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/scalar-data-types/dynamic#operators-and-functions-over-dynamic-types
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/inoperator
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/datatypes-string-operators
        """
        return self.contains(other)

    def assign_to_multiple_columns(self, *columns: 'BaseColumn') -> '_AssignmentToMultipleColumns':
        return _AssignmentToMultipleColumns(columns, self)


@_plain_expression(_KustoType.MAPPING)
class _MappingExpression(_BaseDynamicExpression):
    def __getitem__(self, index: StringType) -> 'AnyExpression':
        return super().__getitem__(index)

    def __getattr__(self, name: str) -> 'AnyExpression':
        return AnyExpression(KQL(f'{self.kql}.{name}'))

    def keys(self) -> _ArrayExpression:
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/bagkeysfunction
        """
        return _ArrayExpression(KQL(f'bag_keys({self.kql})'))

    def bag_contains(self, other: ExpressionType) -> '_BooleanExpression':
        """
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/scalar-data-types/dynamic#operators-and-functions-over-dynamic-types
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/inoperator
        https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/datatypes-string-operators
        """
        return self.contains(other)


class _DynamicExpression(_ArrayExpression, _MappingExpression):
    def __getitem__(self, index: Union[StringType, NumberType]) -> 'AnyExpression':
        return _BaseDynamicExpression.__getitem__(self, index)


class _ComparableExpression(_NumberExpression, _DatetimeExpression, _TimespanExpression):
    # TODO: Implement for strings using 'strcmp': https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/strcmpfunction
    def __lt__(self, other: ComparableType) -> _BooleanExpression:
        return _BooleanExpression.binary_op(self, ' < ', other)

    def __le__(self, other: ComparableType) -> _BooleanExpression:
        return _BooleanExpression.binary_op(self, ' <= ', other)

    def __gt__(self, other: ComparableType) -> _BooleanExpression:
        return _BooleanExpression.binary_op(self, ' > ', other)

    def __ge__(self, other: ComparableType) -> _BooleanExpression:
        return _BooleanExpression.binary_op(self, ' >= ', other)


class AnyExpression(_BooleanExpression, _ComparableExpression, _StringExpression, _DynamicExpression):
    pass


class AggregationExpression(BaseExpression):

    # We would prefer to use 'abc' to make the class abstract, but this can be done only if there is at least one
    # abstract method, which we don't have here. Overriding __new___ is the next best solution.
    def __new__(cls, *args, **kwargs) -> 'AggregationExpression':
        assert cls is not AggregationExpression, "AggregationExpression is abstract"
        return object.__new__(cls)

    def assign_to(self, *columns: '_AnyTypeColumn') -> '_AssignmentFromAggregationToColumn':
        if len(columns) == 0:
            # Unspecified column name
            return _AssignmentFromAggregationToColumn(None, self)
        if len(columns) == 1:
            return _AssignmentFromAggregationToColumn(columns[0], self)
        raise ValueError("Aggregations cannot be assigned to multiple columns")

    def as_subexpression(self) -> KQL:
        return self.kql


@_aggregation_expression(_KustoType.BOOL)
class _BooleanAggregationExpression(AggregationExpression, _BooleanExpression):
    pass


@_aggregation_expression(*_NUMBER_TYPES)
class _NumberAggregationExpression(AggregationExpression, _NumberExpression):
    pass


@_aggregation_expression(_KustoType.STRING)
class _StringAggregationExpression(AggregationExpression, _StringExpression):
    pass


@_aggregation_expression(_KustoType.DATETIME)
class _DatetimeAggregationExpression(AggregationExpression, _DatetimeExpression):
    pass


@_aggregation_expression(_KustoType.TIMESPAN)
class _TimespanAggregationExpression(AggregationExpression, _TimespanExpression):
    pass


@_aggregation_expression(_KustoType.ARRAY)
class _ArrayAggregationExpression(AggregationExpression, _ArrayExpression):
    pass


@_aggregation_expression(_KustoType.MAPPING)
class _MappingAggregationExpression(AggregationExpression, _MappingExpression):
    pass


class _AnyAggregationExpression(AggregationExpression, AnyExpression):
    pass


class _AssignmentBase:
    _lvalue: Optional[KQL]
    _rvalue: KQL

    def __init__(self, lvalue: Optional[KQL], rvalue: ExpressionType) -> None:
        self._lvalue = lvalue
        self._rvalue = _to_kql(rvalue)

    def to_kql(self) -> KQL:
        if self._lvalue is None:
            # Unspecified column name
            return self._rvalue
        return KQL(f'{self._lvalue} = {self._rvalue}')


class _AssignmentToSingleColumn(_AssignmentBase):
    def __init__(self, column: 'BaseColumn', expression: ExpressionType) -> None:
        super().__init__(column.kql, expression)


class _AssignmentFromColumnToColumn(_AssignmentToSingleColumn):
    def __init__(self, target: 'BaseColumn', source: 'BaseColumn') -> None:
        super().__init__(target, source)


class _AssignmentToMultipleColumns(_AssignmentBase):
    def __init__(self, columns: Iterable['BaseColumn'], expression: ArrayType) -> None:
        super().__init__(KQL(f'({", ".join(c.kql for c in columns)})'), expression)


class _AssignmentFromAggregationToColumn(_AssignmentBase):
    def __init__(self, column: Optional['BaseColumn'], aggregation: AggregationExpression) -> None:
        super().__init__(None if column is None else column.kql, aggregation)


class BaseColumn(BaseExpression):
    _name: str

    # We would prefer to use 'abc' to make the class abstract, but this can be done only if there is at least one
    # abstract method, which we don't have here. Overriding __new___ is the next best solution.
    def __new__(cls, *args, **kwargs) -> 'BaseColumn':
        assert cls is not BaseColumn, "BaseColumn is abstract"
        return object.__new__(cls)

    def __init__(self, name: str, quote: bool = False) -> None:
        assert len(name) > 0, "Column name must not be empty"
        should_quote = quote or '.' in name or name in _KUSTO_KEYWORDS or name.isdigit()
        super().__init__(KQL(f"['{name}']" if should_quote else name))
        self._name = name

    def get_name(self) -> str:
        return self._name

    def as_subexpression(self) -> KQL:
        return self.kql

    def assign_to_single_column(self, column: 'BaseColumn') -> '_AssignmentFromColumnToColumn':
        return _AssignmentFromColumnToColumn(column, self)

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self._name})'

    def __str__(self) -> str:
        return self._name


@_typed_column(*_NUMBER_TYPES)
class _NumberColumn(BaseColumn, _NumberExpression):
    pass


@_typed_column(_KustoType.BOOL)
class _BooleanColumn(BaseColumn, _BooleanExpression):
    pass


@_typed_column(_KustoType.ARRAY)
class _ArrayColumn(BaseColumn, _ArrayExpression):
    pass


@_typed_column(_KustoType.MAPPING)
class _MappingColumn(BaseColumn, _MappingExpression):
    pass


class _DynamicColumn(_ArrayColumn, _MappingColumn):
    pass


@_typed_column(_KustoType.STRING)
class _StringColumn(BaseColumn, _StringExpression):
    pass


@_typed_column(_KustoType.DATETIME)
class _DatetimeColumn(BaseColumn, _DatetimeExpression):
    pass


@_typed_column(_KustoType.TIMESPAN)
class _TimespanColumn(BaseColumn, _TimespanExpression):
    pass


class _SubtractableColumn(_NumberColumn, _DatetimeColumn, _TimespanColumn, _ComparableExpression):
    @staticmethod
    def __resolve_type(type_to_resolve: Union['NumberType', 'DatetimeType', 'TimespanType']) -> Optional[_KustoType]:
        # noinspection PyTypeChecker
        base_types = _get_base_types(type_to_resolve)
        possible_types = base_types & ({_KustoType.DATETIME, _KustoType.TIMESPAN} | _NUMBER_TYPES)
        assert len(possible_types) > 0, "Invalid type subtracted"
        if possible_types == _NUMBER_TYPES:
            return _KustoType.INT
        if len(possible_types) > 1:
            return None
        return next(iter(possible_types))

    def __sub__(self, other: Union['NumberType', 'DatetimeType', 'TimespanType']) -> Union['_NumberExpression', '_TimespanExpression', 'AnyExpression']:
        resolved_type = self.__resolve_type(other)
        if resolved_type == _KustoType.DATETIME:
            # Subtracting a datetime can only result in a timespan
            resolved_type = _KustoType.TIMESPAN
        elif resolved_type == _KustoType.TIMESPAN:
            # Subtracting a timespan can result in either a timespan or a datetime
            resolved_type = None
        # Otherwise: subtracting a number can only result in a number

        # noinspection PyTypeChecker
        return BaseExpression.base_binary_op(self, ' - ', other, resolved_type)

    def __rsub__(self, other: Union['NumberType', 'DatetimeType', 'TimespanType']) -> Union['_NumberExpression', '_TimespanExpression', 'AnyExpression']:
        resolved_type = self.__resolve_type(other)
        if resolved_type == _KustoType.DATETIME:
            # Subtracting from a datetime can result in either a datetime or a timespan
            resolved_type = None
        # Otherwise: subtracting from a number or a timespan can only result in a number or a timespan respectively

        # noinspection PyTypeChecker
        return BaseExpression.base_binary_op(other, ' - ', self, resolved_type)


class _AnyTypeColumn(_SubtractableColumn, _BooleanColumn, _DynamicColumn, _StringColumn):
    pass


class ColumnGenerator:
    def __getattr__(self, name: str) -> _AnyTypeColumn:
        return _AnyTypeColumn(name)

    def __getitem__(self, name: str) -> _AnyTypeColumn:
        return _AnyTypeColumn(name)

    # noinspection PyMethodMayBeStatic
    def of(self, name: str) -> _AnyTypeColumn:
        """
        Workaround in case automatic column name quoting fails
        """
        return _AnyTypeColumn(name, quote=True)


# Recommended usage: from pykusto import column_generator as col
# TODO: Is there a way to enforce this to be a singleton?
column_generator = ColumnGenerator()


class _ColumnToType(BaseExpression):
    def __init__(self, col: BaseColumn, kusto_type: _KustoType) -> None:
        super().__init__(KQL(f"{col.kql} to typeof({kusto_type.primary_name})"))


def _to_kql(obj: ExpressionType, parentheses: bool = False) -> KQL:
    """
    Convert the given expression to KQL. If this is a subexpression of a greater expression, neighboring operators might
    take precedence over operators included in this expression, causing an incorrect evaluation order. This might not be a concern, if
    for example the expression is used as an argument to a function.
    If evaluation order is a concern, use `parentheses=True`, which will enclose this expression in parentheses (but only if it is
    a compound expression), thus guaranteeing correct evaluation order.

    :param obj: Expression to convert to KQL
    :return: KQL that represents the given expression
    """
    if isinstance(obj, BaseExpression):
        return obj.as_subexpression() if parentheses else obj.kql
    return _kql_converter.for_obj(obj)


def _expression_to_type(expression: ExpressionType, type_registrar: _TypeRegistrar, fallback_type: Any) -> Any:
    types = set(type_registrar.registry[base_type] for base_type in _plain_expression.get_base_types(expression))
    return next(iter(types)) if len(types) == 1 else fallback_type


_typed_column.assert_all_types_covered()
_plain_expression.assert_all_types_covered()
_aggregation_expression.assert_all_types_covered()



from unittest.mock import patch

import pandas as pd

from pykusto import Query, PySparkKustoClient, ClientRequestProperties
# noinspection PyProtectedMember
from pykusto._src.expressions import _StringColumn, _NumberColumn
# noinspection PyProtectedMember
from test.test_base import TestBase, nested_attribute_dict, captured_stdout


class MockDataFrameReader:
    def __init__(self, dataframe_to_return: pd.DataFrame) -> None:
        self.recorded_format = None
        self.recorded_options = {}
        self.dataframe_to_return = dataframe_to_return

    def format(self, the_format: str) -> 'MockDataFrameReader':
        assert self.recorded_format is None, "Trying to set format twice"
        self.recorded_format = the_format
        return self

    def option(self, key: str, value: str) -> 'MockDataFrameReader':
        assert key not in self.recorded_options, f"Trying to set option '{key}' twice"
        self.recorded_options[key] = value
        return self

    def load(self) -> pd.DataFrame:
        return self.dataframe_to_return


class MockSparkSession:
    def __init__(self, dataframe_to_return: pd.DataFrame) -> None:
        self.read = MockDataFrameReader(dataframe_to_return)


# noinspection PyPep8Naming,PyMethodMayBeStatic
class MockDeviceAuthentication:
    def __init__(self, mock_token: str):
        self.mock_token = mock_token

    def getDeviceCodeMessage(self):
        return "To sign in, use a lubricated goat to open the pod bay doors."

    def acquireToken(self):
        return self.mock_token


class MockSparkContext:
    def __init__(self, mock_token: str):
        self.mock_token = mock_token
        self._jvm = nested_attribute_dict('com.microsoft.kusto.spark.authentication.DeviceAuthentication', lambda s, n, c: MockDeviceAuthentication(self.mock_token))


class TestClient(TestBase):
    def test_linked_service(self):
        rows = (['foo', 10], ['bar', 20], ['baz', 30])
        columns = ('stringField', 'numField')
        expected_df = pd.DataFrame(rows, columns=columns)
        mock_spark_session = MockSparkSession(expected_df)

        with patch('pykusto._src.pyspark_client.PySparkKustoClient._PySparkKustoClient__get_spark_session_and_context', lambda s: (mock_spark_session, None)):
            client = PySparkKustoClient('https://help.kusto.windows.net/', linked_service='MockLinkedKusto', fetch_by_default=False)

        table = client['test_db']['mock_table']
        actual_df = Query(table).take(5).to_dataframe()
        self.assertTrue(expected_df.equals(actual_df))

        self.assertEqual('com.microsoft.kusto.spark.synapse.datasource', mock_spark_session.read.recorded_format)
        self.assertEqual(
            {
                'spark.synapse.linkedService': 'MockLinkedKusto',
                'kustoDatabase': 'test_db',
                'kustoQuery': 'mock_table | take 5',
            },
            mock_spark_session.read.recorded_options,
        )

    def test_linked_service_with_request_properties(self):
        rows = (['foo', 10], ['bar', 20], ['baz', 30])
        columns = ('stringField', 'numField')
        expected_df = pd.DataFrame(rows, columns=columns)
        mock_spark_session = MockSparkSession(expected_df)

        with patch('pykusto._src.pyspark_client.PySparkKustoClient._PySparkKustoClient__get_spark_session_and_context', lambda s: (mock_spark_session, None)):
            client = PySparkKustoClient('https://help.kusto.windows.net/', linked_service='MockLinkedKusto', fetch_by_default=False)

        properties = ClientRequestProperties()
        properties.set_option(ClientRequestProperties.results_defer_partial_query_failures_option_name, False)
        properties.set_parameter('xIntValue', 11)

        table = client['test_db']['mock_table']
        actual_df = Query(table).take(5).to_dataframe(properties=properties)
        self.assertTrue(expected_df.equals(actual_df))

        self.assertEqual('com.microsoft.kusto.spark.synapse.datasource', mock_spark_session.read.recorded_format)
        self.assertEqual(
            {
                'spark.synapse.linkedService': 'MockLinkedKusto',
                'clientRequestPropertiesJson': '{"Options": {"deferpartialqueryfailures": false}, "Parameters": {"xIntValue": 11}}',
                'kustoDatabase': 'test_db',
                'kustoQuery': 'mock_table | take 5',
            },
            mock_spark_session.read.recorded_options,
        )

    def test_linked_service_with_extra_options(self):
        rows = (['foo', 10], ['bar', 20], ['baz', 30])
        columns = ('stringField', 'numField')
        expected_df = pd.DataFrame(rows, columns=columns)
        mock_spark_session = MockSparkSession(expected_df)

        with patch('pykusto._src.pyspark_client.PySparkKustoClient._PySparkKustoClient__get_spark_session_and_context', lambda s: (mock_spark_session, None)):
            client = PySparkKustoClient('https://help.kusto.windows.net/', linked_service='MockLinkedKusto', fetch_by_default=False)

        client.option('alsoMake', 'coffee')
        client.option('performanceLevel', 'awesome')
        client.clear_option('alsoMake')
        table = client['test_db']['mock_table']
        actual_df = Query(table).take(5).to_dataframe()
        self.assertTrue(expected_df.equals(actual_df))

        self.assertEqual('com.microsoft.kusto.spark.synapse.datasource', mock_spark_session.read.recorded_format)
        self.assertEqual(
            {
                'spark.synapse.linkedService': 'MockLinkedKusto',
                'kustoDatabase': 'test_db',
                'kustoQuery': 'mock_table | take 5',
                'performanceLevel': 'awesome',
            },
            mock_spark_session.read.recorded_options,
        )

    def test_device_auth(self):
        rows = (['foo', 10], ['bar', 20], ['baz', 30])
        columns = ('stringField', 'numField')
        expected_df = pd.DataFrame(rows, columns=columns)
        mock_spark_session = MockSparkSession(expected_df)
        mock_spark_context = MockSparkContext('MOCK_TOKEN')

        with patch('pykusto._src.pyspark_client.PySparkKustoClient._PySparkKustoClient__get_spark_session_and_context', lambda s: (mock_spark_session, mock_spark_context)), \
                captured_stdout() as out:
            client = PySparkKustoClient('https://help.kusto.windows.net/', fetch_by_default=False)

        self.assertEqual("To sign in, use a lubricated goat to open the pod bay doors.", out.getvalue().strip())
        table = client['test_db']['mock_table']
        actual_df = Query(table).take(5).to_dataframe()
        self.assertTrue(expected_df.equals(actual_df))

        self.assertEqual('com.microsoft.kusto.spark.datasource', mock_spark_session.read.recorded_format)
        self.assertEqual(
            {
                'kustoCluster': 'https://help.kusto.windows.net/',
                'accessToken': 'MOCK_TOKEN',
                'kustoDatabase': 'test_db',
                'kustoQuery': 'mock_table | take 5',
            },
            mock_spark_session.read.recorded_options,
        )

    def test_repr(self):
        mock_spark_session = MockSparkSession(pd.DataFrame())
        mock_spark_context = MockSparkContext('MOCK_TOKEN')

        with patch('pykusto._src.pyspark_client.PySparkKustoClient._PySparkKustoClient__get_spark_session_and_context', lambda s: (mock_spark_session, mock_spark_context)):
            device_auth_client = PySparkKustoClient('https://help.kusto.windows.net/', fetch_by_default=False)
            linked_service_client = PySparkKustoClient('https://help.kusto.windows.net/', linked_service='MockLinkedKusto', fetch_by_default=False)

        self.assertEqual("PySparkKustoClient('https://help.kusto.windows.net/')", repr(device_auth_client))
        self.assertEqual("PySparkKustoClient('https://help.kusto.windows.net/', 'MockLinkedKusto')", repr(linked_service_client))

    def test_linked_service_with_fetch(self):
        rows = (
            ['test_db', 'mock_table', 'stringField', 'System.String'],
            ['test_db', 'mock_table', 'numField', 'System.Int32'],
        )
        columns = ('DatabaseName', 'TableName', 'ColumnName', 'ColumnType')
        expected_df = pd.DataFrame(rows, columns=columns)
        mock_spark_session = MockSparkSession(expected_df)

        with patch('pykusto._src.pyspark_client.PySparkKustoClient._PySparkKustoClient__get_spark_session_and_context', lambda s: (mock_spark_session, None)):
            client = PySparkKustoClient('https://help.kusto.windows.net/', linked_service='MockLinkedKusto')
            client.wait_for_items()

        self.assertType(client.test_db.mock_table.stringField, _StringColumn)
        self.assertType(client.test_db.mock_table.numField, _NumberColumn)

        self.assertEqual('com.microsoft.kusto.spark.synapse.datasource', mock_spark_session.read.recorded_format)
        self.assertEqual(
            {
                'spark.synapse.linkedService': 'MockLinkedKusto',
                'kustoDatabase': '',
                'kustoQuery': '.show databases schema | project DatabaseName, TableName, ColumnName, ColumnType | limit 100000',
            },
            mock_spark_session.read.recorded_options,
        )



import json
from abc import ABCMeta, abstractmethod
from collections import defaultdict
from fnmatch import fnmatch
from threading import Lock
from typing import Union, List, Tuple, Dict, Generator, Optional, Set, Type, Callable, Iterable

import pandas as pd
from redo import retrier

from .expressions import BaseColumn, _AnyTypeColumn
from .item_fetcher import _ItemFetcher
from .kql_converters import KQL
from .logger import _logger
from .type_utils import _INTERNAL_NAME_TO_TYPE, _typed_column, _DOT_NAME_TO_TYPE, PythonTypes


class RetryConfig:
    def __init__(
            self, attempts: int = 5, sleep_time: float = 60, max_sleep_time: float = 300, sleep_scale: float = 1.5, jitter: float = 1,
            retry_exceptions: Tuple[Type[Exception], ...] = tuple(),
    ) -> None:
        """
        All time parameters are in seconds
        """
        self.attempts = attempts
        self.sleep_time = sleep_time
        self.max_sleep_time = max_sleep_time
        self.sleep_scale = sleep_scale
        self.jitter = jitter
        self.retry_exceptions = retry_exceptions

    def retry(self, action: Callable):
        attempt = 1
        for sleep_time in retrier(attempts=self.attempts, sleeptime=self.sleep_time, max_sleeptime=self.max_sleep_time, sleepscale=self.sleep_scale, jitter=self.jitter):
            try:
                return action()
            except Exception as e:
                for exception_to_check in self.retry_exceptions:
                    if isinstance(e, exception_to_check):
                        if attempt == self.attempts:
                            _logger.warning(f"Reached maximum number of attempts ({self.attempts}), raising exception")
                            raise
                        _logger.info(
                            f"Attempt number {attempt} out of {self.attempts} failed, "
                            f"previous sleep time was {sleep_time} seconds. Exception: {e.__class__.__name__}('{str(e)}')"
                        )
                        break
                else:
                    raise
            attempt += 1

    def retry_on(self, *additional_exceptions: Type[Exception]) -> 'RetryConfig':
        return RetryConfig(self.attempts, self.sleep_time, self.max_sleep_time, self.sleep_scale, self.jitter, self.retry_exceptions + additional_exceptions)


NO_RETRIES = RetryConfig(1)


class KustoResponseBase(metaclass=ABCMeta):
    @abstractmethod
    def get_rows(self) -> Iterable[Iterable]:
        raise NotImplementedError()

    @staticmethod
    def is_row_valid(row: Iterable) -> bool:
        for field in row:
            if field is None or (isinstance(field, str) and len(field.strip()) == 0):
                return False
        return True

    def get_valid_rows(self) -> Generator[Tuple, None, None]:
        for row in self.get_rows():
            if self.is_row_valid(row):
                yield tuple(row)

    @abstractmethod
    def to_dataframe(self) -> pd.DataFrame:
        raise NotImplementedError()


# Copied from https://github.com/Azure/azure-kusto-python/blob/master/azure-kusto-data/azure/kusto/data/client.py
# We are copying this class because we don't won't to force a dependency on azure-kusto-data unless it's actually needed (e.g. it's not needed for PySpark usage).
# noinspection SpellCheckingInspection
class ClientRequestProperties:
    """This class is a POD used by client making requests to describe specific needs from the service executing the requests.
    For more information please look at: https://docs.microsoft.com/en-us/azure/kusto/api/netfx/request-properties
    """

    results_defer_partial_query_failures_option_name = "deferpartialqueryfailures"
    request_timeout_option_name = "servertimeout"
    no_request_timeout_option_name = "norequesttimeout"

    def __init__(self) -> None:
        self._options = {}
        self._parameters = {}
        self.client_request_id = None
        self.application = None
        self.user = None

    @staticmethod
    def _assert_value_is_valid(value: str):
        if not value or not value.strip():
            raise ValueError("Value should not be empty")

    def set_parameter(self, name: str, value: Optional[PythonTypes]):
        """Sets a parameter's value"""
        self._assert_value_is_valid(name)
        self._parameters[name] = value

    def has_parameter(self, name: str) -> bool:
        """Checks if a parameter is specified."""
        return name in self._parameters

    def get_parameter(self, name: str, default_value: Optional[PythonTypes]) -> Optional[PythonTypes]:
        """Gets a parameter's value."""
        return self._parameters.get(name, default_value)

    def set_option(self, name: str, value: Optional[PythonTypes]) -> None:
        """Sets an option's value"""
        self._assert_value_is_valid(name)
        self._options[name] = value

    def has_option(self, name: str) -> bool:
        """Checks if an option is specified."""
        return name in self._options

    def get_option(self, name: str, default_value: Optional[PythonTypes]) -> Optional[PythonTypes]:
        """Gets an option's value."""
        return self._options.get(name, default_value)

    def to_json(self) -> str:
        """Safe serialization to a JSON string."""
        return json.dumps({"Options": self._options, "Parameters": self._parameters}, default=str)


class PyKustoClientBase(_ItemFetcher, metaclass=ABCMeta):
    """
    Handle to a Kusto cluster.
    Uses :class:`ItemFetcher` to fetch and cache the full cluster schema, including all databases, tables, columns and
    their types.
    """

    _cluster_name: str
    _retry_config: RetryConfig
    __first_execution: bool
    __first_execution_lock: Lock

    @abstractmethod
    def __init__(
            self, cluster_name: str, fetch_by_default: bool = True, retry_config: RetryConfig = NO_RETRIES,
    ) -> None:
        """
        Create a new handle to a Kusto cluster. The value of "fetch_by_default" is used for current instance, and also passed on to database instances.

        :param cluster_name: A cluster URL.
        :param retry_config: An instance of RetryConfig which instructs the client how to perform retries in case of failure. The default is NO_RETRIES.
        """
        super().__init__(None, fetch_by_default)
        self._cluster_name = cluster_name
        self._retry_config = retry_config
        self.__first_execution = True
        self.__first_execution_lock = Lock()
        self._refresh_if_needed()

    @abstractmethod
    def __repr__(self) -> str:
        raise NotImplementedError()

    def to_query_format(self) -> KQL:
        return KQL(f'cluster("{self._cluster_name}")')

    def _new_item(self, name: str) -> 'Database':
        # "fetch_by_default" set to false because often a database generated this way is not represented by an actual
        # Kusto database
        return Database(self, name, fetch_by_default=False)

    def get_database(self, name: str) -> 'Database':
        return self[name]

    def execute(self, database: str, query: KQL, properties: ClientRequestProperties = None, retry_config: RetryConfig = None) -> KustoResponseBase:
        # The first execution usually triggers an authentication flow. We block all subsequent executions to prevent redundant authentications.
        # Remove the below block once this is resolved: https://github.com/Azure/azure-kusto-python/issues/208
        with self.__first_execution_lock:
            if self.__first_execution:
                self.__first_execution = False
                return self._internal_execute(database, query, properties, retry_config)
        return self._internal_execute(database, query, properties, retry_config)

    @abstractmethod
    def _internal_execute(self, database: str, query: KQL, properties: ClientRequestProperties = None, retry_config: RetryConfig = None) -> KustoResponseBase:
        raise NotImplementedError()

    def get_databases_names(self) -> Generator[str, None, None]:
        yield from self._get_item_names()

    def get_databases(self) -> Generator['Database', None, None]:
        yield from self._get_items()

    def get_cluster_name(self) -> str:
        return self._cluster_name

    def _internal_get_items(self) -> Dict[str, 'Database']:
        # Retrieves database names, table names, column names and types for all databases. A database name is required
        # by the "execute" method, but is ignored for this query
        res: KustoResponseBase = self.execute(
            '', KQL('.show databases schema | project DatabaseName, TableName, ColumnName, ColumnType | limit 100000')
        )
        database_to_table_to_columns = defaultdict(lambda: defaultdict(list))
        for database_name, table_name, column_name, column_type in res.get_valid_rows():
            database_to_table_to_columns[database_name][table_name].append(
                _typed_column.registry[_DOT_NAME_TO_TYPE[column_type]](column_name)
            )
        return {
            # Database instances are provided with all table and column data, preventing them from generating more
            # queries. However the "fetch_by_default" behavior is passed on to them for future actions.
            database_name: Database(
                self, database_name,
                {table_name: tuple(columns) for table_name, columns in table_to_columns.items()},
                fetch_by_default=self._fetch_by_default
            )
            for database_name, table_to_columns in database_to_table_to_columns.items()
        }


class Database(_ItemFetcher):
    """
    Handle to a Kusto database.
    Uses :class:`ItemFetcher` to fetch and cache the full database schema, including all tables, columns and their
    types.
    """
    __client: PyKustoClientBase
    __name: str

    def __init__(
            self, client: PyKustoClientBase, name: str, tables: Dict[str, Tuple[BaseColumn, ...]] = None,
            fetch_by_default: bool = True
    ) -> None:
        """
        Create a new handle to Kusto database. The value of "fetch_by_default" is used for current instance, and also
        passed on to database instances.

        :param client: The associated PyKustoClient instance
        :param name: Database name
        :param tables: A mapping from table names to the columns of each table. If this is None and "fetch_by_default"
            is true then they will be fetched in the constructor.
        """
        super().__init__(
            # Providing the items to ItemFetcher prevents further queries until the "refresh" method is explicitly
            # called
            None if tables is None else {
                table_name: Table(self, table_name, columns, fetch_by_default=fetch_by_default)
                for table_name, columns in tables.items()
            },
            fetch_by_default
        )
        self.__client = client
        self.__name = name
        self._refresh_if_needed()

    def __repr__(self) -> str:
        return f"{self.__client}.Database('{self.__name}')"

    def to_query_format(self) -> KQL:
        return KQL(f'{self.__client.to_query_format()}.database("{self.__name}")')

    def get_name(self) -> str:
        return self.__name

    def _new_item(self, name: str) -> 'Table':
        # "fetch_by_default" set to false because often a table generated this way is not represented by an actual
        # Kusto table
        return Table(self, name, fetch_by_default=False)

    def execute(self, query: KQL, properties: ClientRequestProperties = None, retry_config: RetryConfig = None) -> KustoResponseBase:
        return self.__client.execute(self.__name, query, properties, retry_config)

    def get_table_names(self) -> Generator[str, None, None]:
        yield from self._get_item_names()

    def get_table(self, *tables: str) -> 'Table':
        assert len(tables) > 0
        if not Table.static_is_union(*tables):
            return self[tables[0]]
        columns: Optional[Tuple[BaseColumn, ...]] = None
        if self._items_fetched():
            resolved_tables: Set[Table] = set()
            for table_pattern in tables:
                if '*' in table_pattern:
                    resolved_tables.update(table for table in self._get_items() if fnmatch(table.get_name(), table_pattern))
                else:
                    resolved_tables.add(self[table_pattern])
            if len(resolved_tables) == 1:
                return next(iter(resolved_tables))
            columns = self.__try_to_resolve_union_columns(*resolved_tables)
        return Table(self, tables, columns, fetch_by_default=self._fetch_by_default)

    @staticmethod
    def __try_to_resolve_union_columns(*resolved_tables: 'Table') -> Optional[Tuple[BaseColumn, ...]]:
        column_by_name: Dict[str, BaseColumn] = {}
        for table in resolved_tables:
            for column in table.get_columns():
                existing_column = column_by_name.setdefault(column.get_name(), column)
                if type(column) is not type(existing_column):
                    return None  # Fallback to Kusto query for column name conflict resolution
        return tuple(column_by_name.values())

    def _internal_get_items(self) -> Dict[str, 'Table']:
        # Retrieves table names, column names and types for this database only (the database name is added in the
        # "execute" method)
        res: KustoResponseBase = self.execute(
            KQL('.show database schema | project TableName, ColumnName, ColumnType | limit 10000')
        )
        table_to_columns = defaultdict(list)
        for table_name, column_name, column_type in res.get_valid_rows():
            table_to_columns[table_name].append(_typed_column.registry[_DOT_NAME_TO_TYPE[column_type]](column_name))
        # Table instances are provided with all column data, preventing them from generating more queries. However the
        # "fetch_by_default" behavior is
        # passed on to them for future actions.
        return {
            table_name: Table(self, table_name, tuple(columns), fetch_by_default=self._fetch_by_default)
            for table_name, columns in table_to_columns.items()
        }


class Table(_ItemFetcher):
    """
    Handle to a Kusto table.
    Uses :class:`ItemFetcher` to fetch and cache the table schema of columns and their types.
    """
    __database: Database
    __tables: Tuple[str, ...]

    def __init__(
            self, database: Database, tables: Union[str, List[str], Tuple[str, ...]],
            columns: Tuple[BaseColumn, ...] = None, fetch_by_default: bool = True
    ) -> None:
        """
        Create a new handle to a Kusto table.

        :param database: The associated Database instance
        :param tables: Either a single table name, or a list of tables. If more than one table is given OR the table
            name contains a wildcard, the Kusto 'union' statement will be used.
        :param columns: Table columns. If this is None and "ItemFetcher" is true then they will be fetched in the
            constructor.
        """
        super().__init__(
            None if columns is None else {c.get_name(): c for c in columns},
            fetch_by_default
        )
        self.__database = database
        self.__tables = (tables,) if isinstance(tables, str) else tuple(tables)
        assert len(self.__tables) > 0
        self._refresh_if_needed()

    def __repr__(self) -> str:
        table_string = ', '.join(f"'{table}'" for table in self.__tables)
        return f'{self.__database}.Table({table_string})'

    def _new_item(self, name: str) -> BaseColumn:
        return _AnyTypeColumn(name)

    def __getattr__(self, name: str) -> BaseColumn:
        """
        Convenience function for obtaining a column using dot notation.
        In contrast with the overridden method from the :class:`ItemFetcher` class, a new column is generated if needed,
        since new columns can be created on the fly in the course of the query (e.g. using 'extend'), and there is no
        fear of undesired erroneous queries sent to Kusto.

        :param name: Name of column
        :return: The column with the given name
        """
        return self[name]

    @staticmethod
    def static_is_union(*table_names: str) -> bool:
        return len(table_names) > 1 or '*' in table_names[0]

    def is_union(self) -> bool:
        return self.static_is_union(*self.__tables)

    def get_name(self) -> str:
        assert not self.is_union()
        return self.__tables[0]

    def to_query_format(self, fully_qualified: bool = False) -> KQL:
        if fully_qualified:
            table_names = tuple(f'{self.__database.to_query_format()}.table("{table}")' for table in self.__tables)
        else:
            table_names = self.__tables
        if self.is_union():
            return KQL('union ' + ', '.join(table_names))
        return KQL(table_names[0])

    def execute(self, query: KQL, properties: ClientRequestProperties = None, retry_config: RetryConfig = None) -> KustoResponseBase:
        return self.__database.execute(query, properties, retry_config)

    def get_columns_names(self) -> Generator[str, None, None]:
        yield from self._get_item_names()

    def get_columns(self) -> Generator[BaseColumn, None, None]:
        yield from self._get_items()

    def _internal_get_items(self) -> Dict[str, BaseColumn]:
        if not self.is_union():
            # Retrieves column names and types for this table only
            res: KustoResponseBase = self.execute(
                KQL(f'.show table {self.get_name()} | project AttributeName, AttributeType | limit 10000')
            )
            return {
                column_name: _typed_column.registry[_INTERNAL_NAME_TO_TYPE[column_type]](column_name)
                for column_name, column_type in res.get_valid_rows()
            }
        # Get Kusto to figure out the schema of the union, especially useful for column name conflict resolution
        res: KustoResponseBase = self.execute(
            KQL(f'{self.to_query_format()} | getschema | project ColumnName, DataType | limit 10000')
        )
        return {
            column_name: _typed_column.registry[_DOT_NAME_TO_TYPE[column_type]](column_name)
            for column_name, column_type in res.get_valid_rows()
        }



import json
from datetime import datetime, timedelta
from itertools import chain
from numbers import Number
from typing import NewType, Union, Mapping, List, Tuple, Any

from .type_utils import _kql_converter, _KustoType, _NUMBER_TYPES

KQL = NewType('KQL', str)


@_kql_converter(_KustoType.DATETIME)
def _datetime_to_kql(dt: datetime) -> KQL:
    return KQL(dt.strftime('datetime(%Y-%m-%d %H:%M:%S.%f)'))


@_kql_converter(_KustoType.TIMESPAN)
def _timedelta_to_kql(td: timedelta) -> KQL:
    hours, remainder = divmod(td.seconds, 3600)
    minutes, seconds = divmod(remainder, 60)
    return KQL(f'time({td.days}.{hours}:{minutes}:{seconds}.{td.microseconds})')


@_kql_converter(_KustoType.ARRAY, _KustoType.MAPPING)
def _dynamic_to_kql(d: Union[Mapping, List, Tuple]) -> KQL:
    try:
        return KQL(f"dynamic({json.dumps(d)})")
    except TypeError:
        # Using exceptions as part of normal flow is not best practice, however in this case we have a good reason.
        # The given object might contain a non-primitive object somewhere inside it, and the only way to find it is to go through the entire hierarchy, which is exactly
        # what's being done in the process of json conversion.
        # Also keep in mind that exception handling in Python has no performance overhead (unlike e.g. Java).
        return _build_dynamic(d)


# Type hint should be Union[PythonTypes, BaseExpression], but currently this is impossible without introducing a circular import.
# Once we drop support for Python 3.6, this enhancement should allow us to use a proper hint: https://www.python.org/dev/peps/pep-0563/
def _build_dynamic(d: Any) -> KQL:
    if isinstance(d, Mapping):
        return KQL(f"pack({', '.join(map(_build_dynamic, chain(*d.items())))})")
    if isinstance(d, (List, Tuple)):
        return KQL(f"pack_array({', '.join(map(_build_dynamic, d))})")
    from .expressions import _to_kql
    return _to_kql(d)


@_kql_converter(_KustoType.BOOL)
def _bool_to_kql(b: bool) -> KQL:
    return KQL('true') if b else KQL('false')


@_kql_converter(_KustoType.STRING)
def _str_to_kql(s: str) -> KQL:
    return KQL(f'"{s}"')


@_kql_converter(*_NUMBER_TYPES)
def _number_to_kql(n: Number) -> KQL:
    return KQL(str(n))


_kql_converter.assert_all_types_covered()



import json
import logging
import sys
from concurrent.futures import Future
from contextlib import contextmanager
from io import StringIO
from threading import Event
from typing import Callable, Tuple, Any, List, Optional, Union, Type
from unittest import TestCase
from urllib.parse import urljoin

from azure.kusto.data import KustoClient, ClientRequestProperties
# noinspection PyProtectedMember
from azure.kusto.data._models import KustoResultTable, KustoResultRow
from azure.kusto.data.response import KustoResponseDataSet

# noinspection PyProtectedMember
from pykusto._src.client import PyKustoClient
# noinspection PyProtectedMember
from pykusto._src.client_base import Table, Database
# noinspection PyProtectedMember
from pykusto._src.expressions import _NumberColumn, _BooleanColumn, _ArrayColumn, _MappingColumn, _StringColumn, _DatetimeColumn, _TimespanColumn, _DynamicColumn
# noinspection PyProtectedMember
from pykusto._src.type_utils import _KustoType

if sys.version_info[1] < 9:
    # noinspection PyProtectedMember
    from unittest.case import _AssertLogsContext
else:
    # noinspection PyUnresolvedReferences,PyProtectedMember,PyCompatibility
    from unittest._log import _AssertLogsContext

# Naming this variable "test_table" triggers the following bug: https://github.com/pytest-dev/pytest/issues/7378
# noinspection PyTypeChecker
mock_table = Table(
    Database(PyKustoClient('mock_cluster', fetch_by_default=False), 'mock_database', fetch_by_default=False), "mock_table",
    (
        _NumberColumn('numField'), _NumberColumn('numField2'), _NumberColumn('numField3'), _NumberColumn('numField4'), _NumberColumn('numField5'), _NumberColumn('numField6'),
        _BooleanColumn('boolField'), _ArrayColumn('arrayField'), _ArrayColumn('arrayField2'), _ArrayColumn('arrayField3'), _MappingColumn('mapField'), _StringColumn('stringField'),
        _StringColumn('stringField2'), _DatetimeColumn('dateField'), _DatetimeColumn('dateField2'), _DatetimeColumn('dateField3'), _TimespanColumn('timespanField'),
        _DynamicColumn('dynamicField')
    )
)


class TestBase(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        logging.basicConfig(
            stream=sys.stdout,
            level=logging.DEBUG,
            format='%(asctime)s %(levelname)5s %(message)s'
        )

    def setUp(self) -> None:
        test_logger.info("Running test: " + self._testMethodName)

    def assertRaises(self, expected_exception: BaseException, test_callable: Callable, *args, **kwargs):
        """
        This method overrides the one in `unittest.case.TestCase`.

        Instead of providing it with an exception type, you provide it with an exception instance that contains
        the expected message.
        """
        expected_exception_type = type(expected_exception)
        expected_exception_message = str(expected_exception)
        with super().assertRaises(expected_exception_type) as cm:
            test_callable(*args, **kwargs)
        self.assertEqual(
            expected_exception_message,
            str(cm.exception)
        )

    def assertLogs(self, logger_to_watch=None, level=None) -> 'CustomAssertLogsContext':
        """
        This method overrides the one in `unittest.case.TestCase`, and has the same behavior, except for not causing a failure when there are no log messages.
        The point is to allow asserting there are no logs.
        Get rid of this in Python 3.10, as this was resolved: https://bugs.python.org/issue39385
        """
        # noinspection PyArgumentList
        return CustomAssertLogsContext(self, logger_to_watch, level)

    @staticmethod
    def raise_mock_exception():
        raise Exception("Mock exception")

    def assertType(self, obj: Any, expected_type: Type):
        self.assertEqual(type(obj), expected_type)


# Get rid of this in Python 3.10, as this was resolved: https://bugs.python.org/issue39385
class CustomAssertLogsContext(_AssertLogsContext):
    # noinspection PyUnresolvedReferences
    def __exit__(self, exc_type, exc_val, exc_tb) -> Optional[bool]:
        # Fool the original exit method to think there is at least one record, to avoid causing a failure
        self.watcher.records.append("DUMMY")
        result = super().__exit__(exc_type, exc_val, exc_tb)
        self.watcher.records.pop()
        return result


# noinspection PyMissingConstructor
class MockKustoResultTable(KustoResultTable):
    def __init__(self, rows: Tuple[Any, ...], columns: Tuple[str, ...]):
        self.kusto_result_rows = tuple(KustoResultRow(columns, row) for row in rows)
        self.raw_rows = self.kusto_result_rows
        self.columns = tuple(type('Column', (object,), {'column_name': col, 'column_type': ''})() for col in columns)


class MockKustoResponseDataSet(KustoResponseDataSet):
    # noinspection PyMissingConstructor
    def __init__(self, rows: Tuple[Any, ...], columns: Tuple[str, ...]):
        self.rows = rows
        self.columns = columns

    @property
    def _error_column(self):
        raise NotImplementedError()

    # noinspection SpellCheckingInspection
    @property
    def _crid_column(self):
        raise NotImplementedError()

    @property
    def _status_column(self):
        raise NotImplementedError()

    @property
    def primary_results(self) -> List[KustoResultTable]:
        return [MockKustoResultTable(self.rows, self.columns)]


# noinspection PyTypeChecker
def mock_response(rows: Tuple[Any, ...], columns: Tuple[str, ...] = tuple()) -> Callable[[], KustoResponseDataSet]:
    return lambda: MockKustoResponseDataSet(rows, columns)


def mock_columns_response(columns: List[Tuple[str, _KustoType]] = tuple()) -> Callable[[], KustoResponseDataSet]:
    return mock_response(tuple((c_name, c_type.internal_name) for c_name, c_type in columns), ('ColumnName', 'ColumnType'))


def mock_tables_response(tables: List[Tuple[str, List[Tuple[str, _KustoType]]]] = tuple()) -> Callable[[], KustoResponseDataSet]:
    return mock_response(
        tuple((t_name, c_name, c_type.dot_net_name) for t_name, columns in tables for c_name, c_type in columns),
        ('TableName', 'ColumnName', 'ColumnType')
    )


def mock_databases_response(databases: List[Tuple[str, List[Tuple[str, List[Tuple[str, _KustoType]]]]]] = tuple()) -> Callable[[], KustoResponseDataSet]:
    return mock_response(
        tuple(
            (d_name, t_name, c_name, c_type.dot_net_name)
            for d_name, tables in databases
            for t_name, columns in tables
            for c_name, c_type in columns
        ),
        ('DatabaseName', 'TableName', 'ColumnName', 'ColumnType')
    )


def mock_getschema_response(columns: List[Tuple[str, _KustoType]] = tuple()) -> Callable[[], KustoResponseDataSet]:
    return mock_response(tuple((c_name, c_type.dot_net_name) for c_name, c_type in columns), ('ColumnName', 'DataType'))


class RecordedQuery:
    database: str
    query: str
    properties: ClientRequestProperties

    def __init__(self, database: str, query: str, properties: ClientRequestProperties = None):
        self.database = database
        self.query = query
        self.properties = properties

    def __repr__(self) -> str:
        return json.dumps({'database': self.database, 'query': self.query, 'properties': self.properties})

    def __eq__(self, o: 'RecordedQuery') -> bool:
        if not isinstance(o, RecordedQuery):
            return False
        return self.database == o.database and self.query == o.query and self.properties == o.properties


# noinspection PyMissingConstructor
class MockKustoClient(KustoClient):
    recorded_queries: List[RecordedQuery]
    columns_response: Callable[[], KustoResponseDataSet]
    tables_response: Callable[[], KustoResponseDataSet]
    databases_response: Callable[[], KustoResponseDataSet]
    getschema_response: Callable[[], KustoResponseDataSet]
    main_response: Callable[[], KustoResponseDataSet]
    record_metadata: bool
    block: bool
    query_future: Union[None, Future]
    blocked_event: Event

    def __init__(
            self,
            cluster="https://test_cluster.kusto.windows.net",
            columns_response: Callable[[], KustoResponseDataSet] = mock_columns_response([('foo', _KustoType.STRING), ('bar', _KustoType.INT)]),
            tables_response: Callable[[], KustoResponseDataSet] = mock_tables_response([
                ('mock_table', [('foo', _KustoType.STRING), ('bar', _KustoType.INT)]),
                ('mock_table_2', [('baz', _KustoType.BOOL)]),
            ]),
            databases_response: Callable[[], KustoResponseDataSet] = mock_databases_response(
                [('test_db', [('mock_table', [('foo', _KustoType.STRING), ('bar', _KustoType.INT)])])]
            ),
            getschema_response: Callable[[], KustoResponseDataSet] = mock_getschema_response([]),
            main_response: Callable[[], KustoResponseDataSet] = mock_response(tuple()),
            record_metadata: bool = False,
            block: bool = False,
    ):
        self.recorded_queries = []
        self._query_endpoint = urljoin(cluster, "/v2/rest/query")
        self.columns_response = columns_response
        self.tables_response = tables_response
        self.databases_response = databases_response
        self.getschema_response = getschema_response
        self.main_response = main_response
        self.record_metadata = record_metadata
        if block:
            self.block = True
            self.query_future = Future()
        else:
            self.block = False
            self.query_future = None
        self.blocked_event = Event()

    def release(self):
        assert self.blocked()
        if self.query_future is not None:
            self.query_future.set_result(None)

    def blocked(self):
        return self.blocked_event.is_set()

    def wait_until_blocked(self):
        self.blocked_event.wait()

    def do_not_block_next_requests(self):
        self.block = False

    def execute(self, database: str, rendered_query: str, properties: ClientRequestProperties = None) -> KustoResponseDataSet:
        recorded_query = RecordedQuery(database, rendered_query, properties)
        if self.block:
            self.blocked_event.set()
            self.query_future.result()
            self.blocked_event.clear()
        metadata_query = True
        if rendered_query == '.show database schema | project TableName, ColumnName, ColumnType | limit 10000':
            response = self.tables_response
        elif rendered_query.startswith('.show table '):
            response = self.columns_response
        elif rendered_query.startswith('.show databases schema '):
            response = self.databases_response
        elif rendered_query.endswith(' | getschema | project ColumnName, DataType | limit 10000'):
            response = self.getschema_response
        else:
            metadata_query = False
            response = self.main_response
        if self.record_metadata or not metadata_query:
            self.recorded_queries.append(recorded_query)
        return response()


def nested_attribute_dict(attributes: str, value: Any) -> Any:
    """
    Used to mock chained lists of attribute references.
    """
    result = value
    for key in reversed(attributes.split('.')):
        result = type(key + 'Wrapper', tuple(), {key: result})()
    return result


@contextmanager
def captured_stdout():
    new_out = StringIO()
    old_out = sys.stdout
    try:
        sys.stdout = new_out
        yield sys.stdout
    finally:
        sys.stdout = old_out


test_logger = logging.getLogger("pykusto_test")



from importlib.util import find_spec
from typing import Dict, Callable, Union, Tuple

import numpy as np
import pandas as pd

from pykusto import NO_RETRIES, KQL, RetryConfig, KustoResponseBase, PyKustoClientBase
from pykusto._src.client_base import ClientRequestProperties


class DataframeBasedKustoResponse(KustoResponseBase):
    """
    In PySpark Kusto results are returned as dataframes. We wrap the dataframe with this object for compatibility with :class:`PyKustoClient`.
    """
    __dataframe: pd.DataFrame

    # noinspection PyMissingConstructor
    def __init__(self, dataframe: pd.DataFrame):
        self.__dataframe = dataframe

    def get_rows(self) -> np.ndarray:
        return self.__dataframe.to_numpy()

    def to_dataframe(self) -> pd.DataFrame:
        return self.__dataframe


class PySparkKustoClient(PyKustoClientBase):
    """
    Handle to a Kusto cluster, to be used inside a PySpark notebook.
    """

    def __init__(self, cluster_name: str, linked_service: str = None, fetch_by_default: bool = True) -> None:
        """
        Create a new handle to a Kusto cluster. The value of "fetch_by_default" is used for current instance, and also passed on to database instances.

        :param cluster_name: a cluster URL.
        :param linked_service: If provided, the connection to Kusto will be made via a pre-configured link (used only for Synapse). Otherwise, device authentication will be used
        (tested only for Synapse, but should work for any PySpark notebook).
        """
        self.__linked_service = linked_service
        super().__init__(cluster_name, fetch_by_default, NO_RETRIES)
        self.__options: Dict[str, Callable[[], str]] = {}
        self.__kusto_session, self.__spark_context = self.__get_spark_session_and_context()

        if self.__linked_service is None:
            # Connect via device authentication
            self.refresh_device_auth()
            self.__format = 'com.microsoft.kusto.spark.datasource'
            self.option('kustoCluster', self._cluster_name)
        else:
            # Connect via pre-configured link
            self.__format = 'com.microsoft.kusto.spark.synapse.datasource'
            self.option('spark.synapse.linkedService', self.__linked_service)

    def __repr__(self) -> str:
        items = [self._cluster_name]
        if self.__linked_service is not None:
            items.append(self.__linked_service)
        item_string = ', '.join(f"'{item}'" for item in items)
        return f"PySparkKustoClient({item_string})"

    def refresh_device_auth(self) -> None:
        """
        Run device authentication sequence, called in the client constructor. Call this method again if you need to re-authenticate.
        """
        assert self.__linked_service is None, "Device authentication can be used only when a linked_service was not provided to the client constructor"
        # noinspection PyProtectedMember
        device_auth = self.__spark_context._jvm.com.microsoft.kusto.spark.authentication.DeviceAuthentication(self._cluster_name, "common")
        print(device_auth.getDeviceCodeMessage())  # Logging is better than printing, but the PySpark notebook does not display logs by default
        self.option('accessToken', device_auth.acquireToken)

    # noinspection PyUnresolvedReferences,PyPackageRequirements
    @staticmethod
    def __get_spark_session_and_context() -> Tuple['pyspark.sql.session.SparkSession', 'pyspark.context.SparkContext']:  # noqa: F821  # pragma: no cover
        if find_spec('pyspark') is None:
            raise RuntimeError("pyspark package not found. PySparkKustoClient can only be used inside a PySpark notebook")
        from pyspark.sql import SparkSession
        from pyspark.context import SparkContext
        return SparkSession.builder.appName("kustoPySpark").getOrCreate(), SparkContext.getOrCreate()

    def option(self, key: str, value: Union[str, Callable[[], str]]) -> 'PySparkKustoClient':
        """
        Add an option to the underlying DataFrameReader. All authentication related options are already handled by this class, but use this method if you need any other options.
        :param key: The option key.
        :param value: Either an option value, or a callable to generate the option value.
        :return: This instance for chained calls.
        """
        if isinstance(value, str):
            self.__options[key] = lambda: value
        else:
            self.__options[key] = value
        return self

    def clear_option(self, key: str) -> 'PySparkKustoClient':
        """
        Clear an option from the underlying DataFrameReader.
        :param key: The option key to clear.
        :return: This instance for chained calls.
        """
        self.__options.pop(key, None)
        return self

    def get_options(self) -> Dict[str, str]:
        """
        Get the options set for the underlying DataFrameReader.
        """
        return {key: value_producer() for key, value_producer in self.__options.items()}

    def _internal_execute(self, database: str, query: KQL, properties: ClientRequestProperties = None, retry_config: RetryConfig = None) -> DataframeBasedKustoResponse:
        resolved_options = self.get_options()
        if properties is not None:
            resolved_options['clientRequestPropertiesJson'] = properties.to_json()
        resolved_options['kustoDatabase'] = database
        resolved_options['kustoQuery'] = query
        kusto_read_session = self.__kusto_session.read.format(self.__format)
        for key, value in resolved_options.items():
            kusto_read_session = kusto_read_session.option(key, value)
        return DataframeBasedKustoResponse(kusto_read_session.load())
